import math

import cv2
import numpy as np
from numpy import linspace
from scipy.interpolate import UnivariateSpline

from ._tailTrackingBase import TailTrackingBase


class CenterOfMassTailTrackingMixin(TailTrackingBase):
  @staticmethod
  def __smoothTail(points, nbTailPoints):
    y = points[0]
    x = linspace(0, 1, len(y))

    if len(x) > 3:
      s = UnivariateSpline(x, y, s=10)
      xs = linspace(0, 1, nbTailPoints)
      newX = s(xs)

      y = points[1]
      x = linspace(0, 1, len(y))
      s = UnivariateSpline(x, y, s=10)
      xs = linspace(0, 1, nbTailPoints)
      newY = s(xs)
    else:
      newX = x
      newY = y

    return [newX, newY]

  def __findNextPoints(self, depth, x, y, frame, points, angle, maxDepth, debug):
    step     = self._hyperparameters["centerOfMassParamStep"]
    segStep  = self._hyperparameters["centerOfMassParamSegStep"]
    halfDiam = self._hyperparameters["centerOfMassParamHalfDiam"]

    if debug:
      frameDisplay = frame.copy()
      frameDisplay = cv2.cvtColor(frameDisplay, cv2.COLOR_GRAY2RGB)

    lenX = len(frame[0]) - 1
    lenY = len(frame) - 1

    thetaDiffAccept = 0.4
    pixTotMax = 1000000
    maxTheta  = angle

    xNew = self._assignValueIfBetweenRange(int(x + step * (math.cos(angle))), 0, lenX)
    yNew = self._assignValueIfBetweenRange(int(y + step * (math.sin(angle))), 0, lenY)

    # if debug:
      # cv2.circle(frameDisplay, (xNew, yNew), 1, (0,0,0),   -1)

    framecopy = frame.copy()

    xmin = self._assignValueIfBetweenRange(xNew - halfDiam, 0, lenX-1)
    xmax = self._assignValueIfBetweenRange(xNew + halfDiam, 0, lenX-1)
    ymin = self._assignValueIfBetweenRange(yNew - halfDiam, 0, lenY-1)
    ymax = self._assignValueIfBetweenRange(yNew + halfDiam, 0, lenY-1)

    if debug:
      cv2.circle(frameDisplay, (xmin, ymin), 1, (255,0,0),   -1)
      cv2.circle(frameDisplay, (xmax, ymax), 1, (255,0,0),   -1)
      cv2.circle(frameDisplay, (x, y),       1, (0,255,0),   -1)

    if (xmin==0 or xmin==lenX-1 or xmax==0 or xmax==lenX-1 or ymin==0 or ymin==lenY-1 or ymax==0 or ymax==lenY-1):
      return (points,0)

    framecopy[0:ymin, :]              = 255
    framecopy[ymax:lenY+1, :]         = 255
    framecopy[:, 0:xmin]              = 255
    framecopy[:, xmax:lenX+1]         = 255

    if debug:
      self._debugFrame(framecopy, title='Frame2')

    (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(framecopy)

    xNew = headPosition[0]
    yNew = headPosition[1]

    theta = self._calculateAngle(x, y, xNew, yNew)

    xNew = self._assignValueIfBetweenRange(int(x + segStep  * (math.cos(theta)) ), 0, lenX)
    yNew = self._assignValueIfBetweenRange(int(y + segStep  * (math.sin(theta)) ), 0, lenY)

    # Calculates distance between new and old point
    distSubsquentPoints = math.sqrt((xNew - x)**2 + (yNew - y)**2)

    if debug:
      cv2.circle(frameDisplay, (xNew, yNew), 1, (0,0,255),   -1)
      self._debugFrame(frameDisplay, title='Frame2')
      self._debugFrame(framecopy, title='Frame2')

    points = self._appendPoint(xNew, yNew, points)
    newTheta = self._calculateAngle(x,y,xNew,yNew)
    if distSubsquentPoints > 0 and depth < maxDepth:
      (points,nop) = self.__findNextPoints(depth+distSubsquentPoints,xNew,yNew,frame,points,newTheta,maxDepth,debug)

    return (points,newTheta)

  def _centerOfMassTailTracking(self, headPosition, frame, maxDepth):
    x = headPosition[0]
    y = headPosition[1]

    initialAngle = self._hyperparameters["headEmbededParamInitialAngle"]

    gaussian_blur = self._hyperparameters["headEmbededParamGaussianBlur"]
    frame = cv2.GaussianBlur(frame, (gaussian_blur, gaussian_blur), 0)
    points = np.zeros((2, 0))

    # if i > 63033:
      # (points, lastFirstTheta2) = self.__findNextPoints(0,x,y,frame,points,initialAngle,maxDepth,True)
    # else:
    (points, lastFirstTheta2) = self.__findNextPoints(0,x,y,frame,points,initialAngle,maxDepth,False)

    output = np.zeros((1, self._nbTailPoints, 2))

    points = np.insert(points, 0, headPosition, axis=1)

    points = points[:,0:(len(points[0])-1)]

    points = self.__smoothTail(points, self._nbTailPoints)

    for idx, x in enumerate(points[0]):
      output[0][idx][0] = x
      output[0][idx][1] = points[1][idx]

    return output

  def _centerOfMassTailTrackFindMaxDepth(self, frame):
    gaussian_blur = self._hyperparameters["headEmbededParamGaussianBlur"]
    frame = cv2.GaussianBlur(frame, (gaussian_blur, gaussian_blur), 0)
    angle = self._hyperparameters["headEmbededParamInitialAngle"]
    points = np.zeros((2, 0))

    x = self._headPositionFirstFrame[0]
    y = self._headPositionFirstFrame[1]

    (points, lastFirstTheta2) = self.__findNextPoints(0,x,y,frame,points,angle,300,False)

    distToTip        = np.full((200),10000)
    curTailLengthTab = np.full((200),10000)
    curTailLength  = 0
    k = 0

    distFromHeadToTip = abs(math.sqrt((x-self._tailTipFirstFrame[0])**2 + (y-self._tailTipFirstFrame[1])**2))
    while (curTailLength < 1.5*distFromHeadToTip) and (k < len(points[0])-1):
      curTailLength = curTailLength + abs(math.sqrt((points[0,k]-points[0,k+1])**2 + (points[1,k]-points[1,k+1])**2))
      distToTip[k]  = abs(math.sqrt((points[0,k]-self._tailTipFirstFrame[0])**2 + (points[1,k]-self._tailTipFirstFrame[1])**2))
      curTailLengthTab[k] = curTailLength
      k = k + 1

    minDistToTip    = 1000000
    indMinDistToTip = 0
    for idx, dist in enumerate(distToTip):
      if dist < minDistToTip:
        minDistToTip = dist
        indMinDistToTip = idx

    return (curTailLengthTab[indMinDistToTip])



import collections
import math
import os
import random
import sys

import cv2
import json
import numpy as np

from PyQt5.QtCore import pyqtSignal, Qt, QSize, QTimer
from PyQt5.QtGui import QColor, QFont, QFontMetrics, QPainter
from PyQt5.QtWidgets import QApplication, QButtonGroup, QCheckBox, QFileDialog, QGridLayout, QHBoxLayout, QMessageBox, QLabel, QProgressDialog, QPushButton, QRadioButton, QSizePolicy, QSlider, QSpinBox, QStyleOptionSlider, QVBoxLayout

import zebrazoom.code.paths as paths
import zebrazoom.code.util as util
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
from zebrazoom.code.createValidationVideo import calculateInfoFrameForFrame, drawInfoFrame, improveContrast
from zebrazoom.code.getHyperparameters import getHyperparametersSimple
from zebrazoom.code.preprocessImage import preprocessImage
from zebrazoom.code.GUI.adjustParameterInsideAlgo import adjustFastFishTrackingPage


class _CustomTimer(QTimer):
  started = pyqtSignal()
  stopped = pyqtSignal()

  def start(self):
    super().start()
    self.started.emit()

  def stop(self):
    super().stop()
    self.stopped.emit()


class FrameSlider(QSlider):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.firstSaveFrame = None

  def paintEvent(self, event):
    if self.firstSaveFrame is not None:
      style = self.style()
      opt = QStyleOptionSlider()
      self.initStyleOption(opt)
      available = style.pixelMetric(style.PM_SliderSpaceAvailable, opt, self)
      minimum = self.minimum()
      maximum = self.maximum()
      qp = QPainter(self)
      qp.translate(opt.rect.x(), opt.rect.y())
      start = style.sliderPositionFromValue(minimum, maximum, self.firstSaveFrame, available, opt.upsideDown)
      end = style.sliderPositionFromValue(minimum, maximum, self.value(), available, opt.upsideDown)
      qp.fillRect(start, 0, end - start + 1, event.rect().height(), QColor('red'))
    super().paintEvent(event)


def getFramesCallback(videoPath, folderName, numWell, numAnimal, zoom, start, framesToShow=0, ZZoutputLocation='', supstruct=None, config=None):
  s1  = "ZZoutput"
  s2  = folderName
  s3b = "results_"
  s4  = folderName
  s5  = ".avi"
  s5b = ".txt"

  if len(ZZoutputLocation):
    initialPath = ZZoutputLocation
  else:
    initialPath = paths.getDefaultZZoutputFolder()

  if config is None:
    with open(os.path.join(initialPath, os.path.join(s2, 'configUsed.json'))) as f:
      config = json.load(f)
  hyperparameters = getHyperparametersSimple(config)

  if os.path.splitext(folderName)[1] != '.h5':
    folderPath = os.path.join(initialPath, folderName)
    resultFile = next((f for f in os.listdir(folderPath) if os.path.isfile(os.path.join(folderPath, f)) and os.path.splitext(f)[1] == '.txt' and f.startswith('results_')), None)
    if resultFile is None:
      return
    resultsPath = os.path.join(initialPath, folderName, resultFile)
    if supstruct is None:
      with open(resultsPath) as f:
        supstruct = json.load(f)
  else:
    resultsPath = os.path.join(initialPath, folderName)
  
  if "pathToOriginalVideo" in supstruct:
    videoPath = supstruct["pathToOriginalVideo"]
    if not os.path.exists(videoPath):
      app = QApplication.instance()
      if QMessageBox.critical(app.window, "Video not found", "Cannot display the validation video because the video used to run the tracking can no longer be found. Would you like to update the video path stored in the results file?",
                              buttons=QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No, defaultButton=QMessageBox.StandardButton.Yes) != QMessageBox.StandardButton.Yes:
        return None
      videoName, _ = QFileDialog.getOpenFileName(app.window, 'Select video', os.path.expanduser("~"))
      if not videoName:
        return
      videoPath = supstruct["pathToOriginalVideo"] = videoName
      if os.path.splitext(resultsPath)[1] != '.h5':
        with open(resultsPath, 'w') as f:
          json.dump(supstruct, f)
      else:
        import h5py
        with h5py.File(resultsPath, 'a') as results:
          results.attrs['pathToOriginalVideo'] = videoName
  else:
    if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] and os.path.exists(os.path.join(initialPath, os.path.join(s1, os.path.join(s2, 'originalVideoWithoutAnyTrackingDisplayed_pleaseUseTheGUIToVisualizeTrackingPoints.avi')))):
      # The "exist" check above is only to insure compatibility with videos tracked prior to this update
      if os.path.splitext(resultsPath)[1] != '.h5':
        videoPath = os.path.join(initialPath, os.path.join(s2, 'originalVideoWithoutAnyTrackingDisplayed_pleaseUseTheGUIToVisualizeTrackingPoints.avi'))
      else:
        videoPath = f'{os.path.splitext(resultsPath)[0]}_originalVideoWithoutAnyTrackingDisplayed_pleaseUseTheGUIToVisualizeTrackingPoints.avi'
    else:
      if os.path.splitext(resultsPath)[1] != '.h5':
        videoPath = os.path.join(initialPath, os.path.join(s2, s4 + s5))
      else:
        videoPath = f'{os.path.splitext(resultsPath)[0]}.avi'

  if not(os.path.exists(videoPath)) and os.path.splitext(resultsPath)[1] != '.h5':
    mypath = os.path.join(initialPath, s2)
    onlyfiles = [f for f in os.listdir(mypath) if os.path.isfile(os.path.join(mypath, f))]
    resultFile = ''
    for fileName in onlyfiles:
      if '.avi' in fileName:
        resultFile = fileName
    videoPath = os.path.join(initialPath, os.path.join(s2, resultFile))

  if not os.path.isfile(videoPath):
    app = QApplication.instance()
    QMessageBox.critical(app.window, "Video not found", "Cannot display the validation video because it could not be found.")
    return

  cap = zzVideoReading.VideoCapture(videoPath, hyperparameters)

  nx    = int(cap.get(3))
  ny    = int(cap.get(4))
  max_l = int(cap.get(7)) if int(cap.get(7)) != -1 else hyperparameters["lastFrame"] # The "if" is to deal with eventBased data reading
  if max_l == 1:
    return None

  if not("firstFrame" in supstruct):
    supstruct["firstFrame"] = 1
    print("supstruct['firstFrame'] not found")

  if "lastFrame" not in supstruct:
    supstruct["lastFrame"] = max_l - 1
    print("supstruct['lastFrame'] not found")

  frameRange = (supstruct["firstFrame"], supstruct["lastFrame"] - 1)

  infoWells = []

  HeadX = np.zeros(max_l)
  HeadY = np.zeros(max_l)

  if ((numWell != -1) and (zoom)):
    lastEnd = 0
    lastXpos = supstruct["wellPoissMouv"][numWell][numAnimal][0]["HeadX"][0]
    lastYpos = supstruct["wellPoissMouv"][numWell][numAnimal][0]["HeadY"][0]
    for bout in supstruct["wellPoissMouv"][numWell][numAnimal]:
      beg = bout["BoutStart"]
      end = bout["BoutEnd"]
      if not (hyperparameters["copyOriginalVideoToOutputFolderForValidation"] or "pathToOriginalVideo" in supstruct):
        beg -= supstruct["firstFrame"]
        end -= supstruct["firstFrame"]
      HeadX[lastEnd:beg] = lastXpos
      HeadX[beg:end+1] = bout['HeadX'][:None if end + 1 <= max_l else max_l - beg]
      HeadY[lastEnd:beg] = lastYpos
      HeadY[beg:end+1] = bout['HeadY'][:None if end + 1 <= max_l else max_l - beg]
      lastEnd = end + 1
      lastXpos = bout["HeadX"][end-1-beg]
      lastYpos = bout["HeadY"][end-1-beg]

    HeadX[lastEnd:] = lastXpos
    HeadY[lastEnd:] = lastYpos

  # /* Getting the info about well positions */
  analyzeAllWellsAtTheSameTime = 0
  if (analyzeAllWellsAtTheSameTime == 0):
    for i in range(0, len(supstruct["wellPositions"])):
      x = 0
      y = 0
      lengthX = 0
      lengthY = 0
      rectangleWellArea = 1
      if (rectangleWellArea == 0): # circular wells
        x = supstruct["wellPositions"][i]["topLeftX"]
        y = supstruct["wellPositions"][i]["topLeftY"]
        r = supstruct["wellPositions"][i]["diameter"]
        lengthX = 300 # wellOutputVideoDiameter;
        lengthY = 300 # wellOutputVideoDiameter;
      else:
        x = supstruct["wellPositions"][i]["topLeftX"]
        y = supstruct["wellPositions"][i]["topLeftY"]
        lengthX = supstruct["wellPositions"][i]["lengthX"]
        lengthY = supstruct["wellPositions"][i]["lengthY"]
      if (x < 0):
        x = 0
      if (y < 0):
        y = 0
      infoWells.append([x, y, lengthX, lengthY])
  else:
    infoWells.append([0, 0, nx, ny])
  
  boutMap = None
  if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] or "pathToOriginalVideo" in supstruct:
    boutMap = collections.defaultdict(list)
    for wellIdx, well in enumerate(supstruct["wellPoissMouv"]):
      for animalIdx, animal in enumerate(well):
        for boutIdx, bout in enumerate(animal):
          for frame in range(bout['BoutStart'], bout['BoutEnd'] + 1):
            boutMap[frame].append((wellIdx, animalIdx, boutIdx))
    colorModifTab = [{"red": random.randrange(255), "green": random.randrange(255), "blue": random.randrange(255)} for i in range(1, hyperparameters["nbAnimalsPerWell"])]
    colorModifTab.insert(0, {"red": 0, "green": 0, "blue": 0})

  x = 0
  y = 0
  lengthX = 0
  lengthY = 0
  if (numWell != -1):
    x = infoWells[numWell][0]
    y = infoWells[numWell][1]
    lengthX = infoWells[numWell][2]
    lengthY = infoWells[numWell][3]
  else:
    lengthX = nx
    lengthY = ny

  xOriginal = x
  yOriginal = y

  def getFrame(frameSlider, timer=None, trackingPointsGroup=None, stopTimer=True, returnHeadPos=False, returnOffsets=False, frameIdx=None, returnFPS=False, topLeftCircleCb=None):
    nonlocal x
    nonlocal y
    nonlocal lengthX
    nonlocal lengthY

    l = frameSlider.value() if frameIdx is None else frameIdx
    if timer is not None and timer.isActive() and (l == frameSlider.maximum() or stopTimer):
      timer.stop()

    if not (hyperparameters["copyOriginalVideoToOutputFolderForValidation"] or "pathToOriginalVideo" in supstruct):
      l -= supstruct['firstFrame']

    cap.set(1, l)
    ret, img = cap.read()
    if type(img[0][0]) != np.ndarray:
      img = cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
    
    if hyperparameters["imagePreProcessMethod"]:
      img = preprocessImage(img, hyperparameters)
    if hyperparameters["outputValidationVideoContrastImprovement"]:
      img = improveContrast(img, hyperparameters)

    if boutMap is not None and l in boutMap and trackingPointsGroup is not None and trackingPointsGroup.checkedId():
      hyperparameters["plotOnlyOneTailPointForVisu"] = trackingPointsGroup.checkedId() == 1
      infoFrame = [info for args in boutMap[l] for info in calculateInfoFrameForFrame(supstruct, hyperparameters, *args, l, colorModifTab)]
      drawInfoFrame(img, infoFrame, colorModifTab, hyperparameters)

    if numWell != -1 and zoom:
      length = 250
      xmin = int(HeadX[l] - length/2)
      xmax = int(HeadX[l] + length/2)
      ymin = int(HeadY[l] - length/2)
      ymax = int(HeadY[l] + length/2)

      x = max(xmin + xOriginal, 0)
      y = max(ymin + yOriginal, 0)
      if returnHeadPos:
        return HeadX[l] - x + xOriginal, HeadY[l] - y + yOriginal
      if returnOffsets:
        return x - xOriginal, y - yOriginal
      lengthX = xmax - xmin
      lengthY = ymax - ymin

      if y + lengthY >= len(img):
        lengthY = len(img) - y - 1
      if x + lengthX >= len(img[0]):
        lengthX = len(img[0]) - x - 1

    if (numWell != -1):
      img = img[y:y+lengthY, x:x+lengthX]

    frameNumber = l if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] or "pathToOriginalVideo" in supstruct else (l + supstruct["firstFrame"])
    if lengthX > 100 and lengthY > 100:
      font = cv2.FONT_HERSHEY_SIMPLEX
      cv2.putText(img,str(frameNumber),(int(lengthX-110), int(lengthY-30)),font,1,(0,255,0))
    else:
      blank_image = np.zeros((len(img)+30, len(img[0]), 3), np.uint8)
      blank_image[0:len(img), 0:len(img[0])] = img
      img = blank_image
      font = cv2.FONT_HERSHEY_SIMPLEX
      cv2.putText(img, str(frameNumber), (int(0), int(lengthY+25)), font, 1, (0,255,0))

    if trackingPointsGroup is not None and not trackingPointsGroup.checkedId() and topLeftCircleCb is not None and topLeftCircleCb():
      cv2.circle(img, (20, 20), 20, (0, 0, 255), -1)

    return (img, int(cap.get(5) if not hyperparameters['outputValidationVideoFps'] > 0 else hyperparameters['outputValidationVideoFps'])) if returnFPS else img

  wellShape = None if config.get("noWellDetection", False) or (hyperparameters["headEmbeded"] and not hyperparameters["oneWellManuallyChosenTopLeft"]) else 'rectangle' if config.get("wellsAreRectangles", False) or len(config.get("oneWellManuallyChosenTopLeft", '')) or int(config.get("multipleROIsDefinedDuringExecution", 0)) or config.get("groupOfMultipleSameSizeAndShapeEquallySpacedWells", False) else 'circle'
  return getFrame, frameRange, start if start > 0 else 0, boutMap is not None, supstruct['wellPositions'], wellShape, hyperparameters, videoPath if hyperparameters["copyOriginalVideoToOutputFolderForValidation"] or "pathToOriginalVideo" in supstruct else None


class _ElidedLabel(QLabel):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self._value = None
    self._fm = QFontMetrics(QFont())

  def setValue(self, value, formatCb):
    self._value = value
    self.setText('' if value is None else self._fm.elidedText(formatCb(value), Qt.TextElideMode.ElideRight, self.size().width()))

  def getValue(self):
    return self._value


def readValidationVideo(videoPath, folderName, numWell, numAnimal, zoom, start, framesToShow=0, ZZoutputLocation='', supstruct=None, config=None):
  frameInfo = getFramesCallback(videoPath, folderName, numWell, numAnimal, zoom, start, framesToShow=framesToShow, ZZoutputLocation=ZZoutputLocation, supstruct=supstruct, config=config)
  if frameInfo is None:
    return
  getFrame, frameRange, frame, toggleTrackingPoints, wellPositions, wellShape, hyperparameters, originalVideoPath = frameInfo
  if wellShape is None:
    wellShape = 'rectangle'
  layout = QVBoxLayout()

  video = QLabel()
  if zoom:
    video.setAlignment(Qt.AlignmentFlag.AlignCenter)
    video.minimumSizeHint = lambda: QSize(250, 250)
  sizePolicy = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
  sizePolicy.setRetainSizeWhenHidden(True)
  video.setSizePolicy(sizePolicy)
  topLeftCircleCb = None
  if zoom and folderName.endswith('.h5'):
    videoLayout = QHBoxLayout()
    videoLayout.addWidget(video, stretch=1)
    layout.addLayout(videoLayout, stretch=1)
    validationLayout = QGridLayout()
    validationLayout.setColumnStretch(2, 1)
    validationLayout.setRowStretch(5, 1)
    bendCheckbox = QCheckBox('Bend')
    bendCheckbox.toggled.connect(lambda checked: saveChanges() or clearButton.setEnabled(True) or util.setPixmapFromCv(getFrame(frameSlider, timer, btnGroup, stopTimer, topLeftCircleCb=topLeftCircleCb), video))
    validationLayout.addWidget(bendCheckbox, 1, 0, 1, 2)
    headingButton = QPushButton('Heading:')

    def updateHeading():
      from zebrazoom.code.extractParameters import calculateAngle
      exitSignals = [frameSlider.valueChanged, manualValidationWidget.expanded, headingButton.clicked, tailExtremityButton.clicked]
      if toggleTrackingPoints:
        exitSignals.extend([btnGroup.idToggled, sizeSpinbox.valueChanged, contrastCheckbox.toggled])
      points = util.getPointOnFrame(getFrame(frameSlider, timer, btnGroup, stopTimer), video, basePoint=getFrame(frameSlider, returnHeadPos=True), exitSignals=exitSignals)
      if points is None:
        return
      headingLabel.setValue(calculateAngle(*points), str)
      saveChanges()
      clearButton.setEnabled(True)

    headingButton.clicked.connect(lambda: QTimer.singleShot(0, updateHeading))
    validationLayout.addWidget(headingButton, 2, 0)
    headingLabel = _ElidedLabel()
    validationLayout.addWidget(headingLabel, 2, 1, 1, 2)
    tailExtremityButton = QPushButton('Tail extremity:')

    def updateTailExtremity():
      exitSignals = [frameSlider.valueChanged, manualValidationWidget.expanded, headingButton.clicked, tailExtremityButton.clicked]
      if toggleTrackingPoints:
        exitSignals.extend([btnGroup.idToggled, sizeSpinbox.valueChanged, contrastCheckbox.toggled])
      point = util.getPointOnFrame(getFrame(frameSlider, timer, btnGroup, stopTimer), video, exitSignals=exitSignals)
      if point is None:
        return
      tailExtremityLabel.setValue(tuple(map(sum, zip(point, getFrame(frameSlider, returnOffsets=True)))), lambda point: ', '.join(map(str, point)))
      saveChanges()
      clearButton.setEnabled(True)

    tailExtremityButton.clicked.connect(lambda: QTimer.singleShot(0, updateTailExtremity))
    validationLayout.addWidget(tailExtremityButton, 3, 0)
    tailExtremityLabel = _ElidedLabel()
    validationLayout.addWidget(tailExtremityLabel, 3, 1, 1, 2)
    clearButton = QPushButton('Clear values')
    clearButton.setEnabled(False)
    clearButton.clicked.connect(lambda: updateWidgets(False, np.nan, (-1, -1)) or saveChanges() or clearButton.setEnabled(False) or util.setPixmapFromCv(getFrame(frameSlider, timer, btnGroup, stopTimer, topLeftCircleCb=topLeftCircleCb), video))
    validationLayout.addWidget(clearButton, 4, 0, 1, 3, alignment=Qt.AlignmentFlag.AlignCenter)

    def saveChanges():
      import h5py
      filePath = os.path.join(ZZoutputLocation if ZZoutputLocation else paths.getDefaultZZoutputFolder(), folderName)
      with h5py.File(filePath, 'r+') as results:
        firstFrame = results.attrs['firstFrame']
        arraySize = results.attrs['lastFrame'] - firstFrame + 1
        perFrameGroup = results[f'dataForWell{numWell}'][f'dataForAnimal{numAnimal}']['dataPerFrame']
        if 'manualBend' not in perFrameGroup:
          perFrameGroup.create_dataset('manualBend', data=np.full(arraySize, False, dtype=bool))
        if 'manualHeading' not in perFrameGroup:
          perFrameGroup.create_dataset('manualHeading', data=np.full(arraySize, np.nan))
        if 'manualTailExtremity' not in perFrameGroup:
          perFrameGroup.create_dataset('manualTailExtremity', data=np.full(arraySize, -1, dtype=[('X', int), ('Y', int)]))  # float array is required in order to support nan
          perFrameGroup['manualTailExtremity'].attrs['columns'] = ('X', 'Y')
        idx = frameSlider.value() - firstFrame
        bend = bendCheckbox.isChecked()
        perFrameGroup['manualBend'][idx] = bend if bend != -1 else np.nan
        heading = headingLabel.getValue()
        perFrameGroup['manualHeading'][idx] = heading if heading is not None else np.nan
        tailExtremity = tailExtremityLabel.getValue()
        perFrameGroup['manualTailExtremity'][idx] = tailExtremity if tailExtremity is not None else (-1, -1)

    manualValidationWidget = util.Expander(None, 'Manual validation', validationLayout, retainWidth=True)

    def updateWidgets(bend, heading, tailExtremity):
      blocked = bendCheckbox.blockSignals(True)
      bendCheckbox.setChecked(bend)
      bendCheckbox.blockSignals(blocked)
      headingLabel.setValue(heading if not np.isnan(heading) else None, str)
      tailExtremityLabel.setValue(tailExtremity if tailExtremity != (-1, -1) else None, lambda tailExtremity: ', '.join(map(str, tailExtremity)))

    def manualValidationExpanded(expanded):
      timer.stop()
      if not expanded:
        util.setPixmapFromCv(getFrame(frameSlider, timer, btnGroup, stopTimer, topLeftCircleCb=topLeftCircleCb), video)
        return
      import h5py
      filePath = os.path.join(ZZoutputLocation if ZZoutputLocation else paths.getDefaultZZoutputFolder(), folderName)
      with h5py.File(filePath, 'r+') as results:
        idx = frameSlider.value() - results.attrs['firstFrame']
        perFrameGroup = results[f'dataForWell{numWell}'][f'dataForAnimal{numAnimal}']['dataPerFrame']
        bend = perFrameGroup['manualBend'][idx] if 'manualBend' in perFrameGroup else False
        heading = perFrameGroup['manualHeading'][idx] if 'manualHeading' in perFrameGroup else np.nan
        tailExtremity = tuple(perFrameGroup['manualTailExtremity'][idx]) if 'manualTailExtremity' in perFrameGroup else (-1, -1)
      updateWidgets(bend, heading, tailExtremity)
      clearButton.setEnabled(bool(tailExtremityLabel.text() or headingLabel.text() or bendCheckbox.isChecked()))
      util.setPixmapFromCv(getFrame(frameSlider, timer, btnGroup, stopTimer, topLeftCircleCb=topLeftCircleCb), video)

    manualValidationWidget.expanded.connect(manualValidationExpanded)
    videoLayout.addWidget(manualValidationWidget, alignment=Qt.AlignmentFlag.AlignTop)
    topLeftCircleCb = lambda: manualValidationWidget.isExpanded() and bendCheckbox.isChecked()
  else:
    layout.addWidget(video, stretch=1)

  sliderLayout = QHBoxLayout()
  frameSlider = FrameSlider(Qt.Orientation.Horizontal)
  frameSlider.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
  frameSlider.setPageStep(50)
  frameSlider.setRange(*frameRange)
  frameSlider.setValue(frame)

  def updateManualValidationWidgets(value):
    blocked = frameSpinbox.blockSignals(True)
    frameSpinbox.setValue(frameSlider.value())
    frameSpinbox.blockSignals(blocked)
    if zoom and folderName.endswith('.h5'):
      if not manualValidationWidget.isExpanded():
        return
      manualValidationExpanded(True)
  frameSlider.valueChanged.connect(updateManualValidationWidgets)

  frameSlider.valueChanged.connect(lambda: util.setPixmapFromCv(getFrame(frameSlider, timer, btnGroup, stopTimer, topLeftCircleCb=topLeftCircleCb), video))
  sliderLayout.addWidget(frameSlider)

  frameSpinbox = QSpinBox()
  frameSpinbox.setRange(*frameRange)
  frameSpinbox.setStyleSheet(util.SPINBOX_STYLESHEET)
  frameSpinbox.setMinimumWidth(70)
  frameSpinbox.valueChanged.connect(lambda value: timer.stop() or frameSlider.setValue(value))
  sliderLayout.addWidget(frameSpinbox)

  playBtn = QPushButton()
  style = playBtn.style()
  playIcon = style.standardIcon(style.SP_MediaPlay)
  pauseIcon = style.standardIcon(style.SP_MediaPause)
  playBtn.setIcon(playIcon)
  playBtn.clicked.connect(lambda: timer.stop() if timer.isActive() else timer.start())
  sliderLayout.addWidget(playBtn)

  layout.addLayout(sliderLayout)
  shortcutsLabel = QLabel("Left Arrow, Right Arrow, Page Up, Page Down, Home and End keys can be used to navigate through the video.")
  shortcutsLabel.setWordWrap(True)
  layout.addWidget(shortcutsLabel)
  btnGroup = None
  trackingPointsLayout = QHBoxLayout()
  if toggleTrackingPoints:
    plotTrackingPointsLabel = QLabel("Tracking points:")
    trackingPointsLayout.addWidget(plotTrackingPointsLabel)
    btnGroup = QButtonGroup()
    allPointsRadioButton = QRadioButton('All')
    allPointsRadioButton.setChecked(True)
    trackingPointsLayout.addWidget(allPointsRadioButton)
    btnGroup.addButton(allPointsRadioButton, id=2)
    minimalPointsRadioButton = QRadioButton('Minimal')
    minimalPointsRadioButton.setChecked(bool(hyperparameters['plotOnlyOneTailPointForVisu']))
    trackingPointsLayout.addWidget(minimalPointsRadioButton)
    btnGroup.addButton(minimalPointsRadioButton, id=1)
    noPointsRadioButton = QRadioButton('None')
    trackingPointsLayout.addWidget(noPointsRadioButton)
    btnGroup.addButton(noPointsRadioButton, id=0)
    btnGroup.idToggled.connect(lambda: timer.isActive() or util.setPixmapFromCv(getFrame(frameSlider, timer, btnGroup, stopTimer, topLeftCircleCb=topLeftCircleCb), video))
    trackingPointsLayout.addWidget(QLabel('Size:'))
    sizeSpinbox = QSpinBox()
    sizeSpinbox.setMinimum(1)
    sizeSpinbox.setValue(hyperparameters["trackingPointSizeDisplay"])
    sizeSpinbox.valueChanged.connect(lambda value: hyperparameters.update({"trackingPointSizeDisplay": value}))
    sizeSpinbox.valueChanged.connect(lambda: timer.isActive() or util.setPixmapFromCv(getFrame(frameSlider, timer, btnGroup, stopTimer, topLeftCircleCb=topLeftCircleCb), video))
    trackingPointsLayout.addWidget(sizeSpinbox)
    contrastCheckbox = QCheckBox('Contrast')
    contrastCheckbox.setChecked(bool(hyperparameters['outputValidationVideoContrastImprovement']))
    contrastCheckbox.toggled.connect(lambda checked: hyperparameters.update({"outputValidationVideoContrastImprovement": int(checked)}))
    contrastCheckbox.toggled.connect(lambda: timer.isActive() or util.setPixmapFromCv(getFrame(frameSlider, timer, btnGroup, stopTimer, topLeftCircleCb=topLeftCircleCb), video))
    trackingPointsLayout.addWidget(contrastCheckbox)
    saveButton = QPushButton('Start save')

    def saveVideo():
      if frameSlider.firstSaveFrame is None:
        frameSlider.firstSaveFrame = frameSlider.value()
        timer.stop()
        saveButton.setText('End save')
        return

      app = QApplication.instance()
      filename, _ = QFileDialog.getSaveFileName(app.window, 'Select file', os.path.expanduser('~'), "AVI (*.avi)")
      if filename:
        lastSaveFrame = frameSlider.value()
        if lastSaveFrame < frameSlider.firstSaveFrame:
          frameSlider.firstSaveFrame, lastSaveFrame = lastSaveFrame, frameSlider.firstSaveFrame
        progressDialog = QProgressDialog("Saving video...", "Cancel", frameSlider.firstSaveFrame, lastSaveFrame, app.window, Qt.WindowType.WindowTitleHint | Qt.WindowType.WindowCloseButtonHint)
        progressDialog.setWindowTitle('Saving Video')
        progressDialog.setAutoClose(False)
        progressDialog.setAutoReset(False)
        progressDialog.setWindowModality(Qt.WindowModality.ApplicationModal)
        progressDialog.setMinimumDuration(0)
        frame, videoFPS = getFrame(frameSlider, trackingPointsGroup=btnGroup, frameIdx=frameSlider.firstSaveFrame, returnFPS=True)
        height, width = frame.shape[:2] if not zoom else (250, 250)
        cap = cv2.VideoWriter(filename, cv2.VideoWriter_fourcc(*'MJPG'), videoFPS, (width, height))
        for frameIdx in range(frameSlider.firstSaveFrame, lastSaveFrame + 1):
          progressDialog.setValue(frameIdx)
          progressDialog.setLabelText(f'Saving frame {frameIdx}...')
          frame = getFrame(frameSlider, trackingPointsGroup=btnGroup, frameIdx=frameIdx)
          frameHeight, frameWidth = frame.shape[:2]
          if frameHeight < height or frameWidth < width:
            xOffset = (width - frameWidth) // 2
            yOffset = (height - frameHeight) // 2
            copy = cv2.resize(frame, (height, width))
            copy[:] = 0
            copy[yOffset:yOffset+frameHeight, xOffset:xOffset+frameWidth] = frame
            frame = copy
          cap.write(frame)
          if progressDialog.wasCanceled():
            break
        progressDialog.setLabelText('Saving video...')
        cap.release()
        progressDialog.close()
      saveButton.setText('Start save')
      frameSlider.firstSaveFrame = None

    saveButton.clicked.connect(saveVideo)
    trackingPointsLayout.addWidget(saveButton)

  adjustButton = None
  if config is not None and (numWell != -1 or config.get("headEmbeded", False)) and ('trackingImplementation' not in config or config['trackingImplementation'] == 'fastFishTracking.tracking') and not config.get("trackingMethod"):
    adjustButton = QPushButton('Adjust parameters')
    def adjustParametrsClicked():
      app = QApplication.instance()
      app.configFile = {key: val for key, val in config.items() if key != 'popUpAlgoFollow'}
      app.savedConfigFile = app.configFile.copy()
      if originalVideoPath is None:
        app.videoToCreateConfigFileFor, _ = QFileDialog.getOpenFileName(app.window, "Select video to create config file for", os.path.expanduser("~"),
                                                                        filter=f'Videos ({" ".join("*%s" % ext for ext in util.VIDEO_EXTENSIONS)});; All files (*.*)')
      else:
        app.videoToCreateConfigFileFor = originalVideoPath.replace('\\', '/')
      if not app.videoToCreateConfigFileFor:
        app.configFile.clear()
        return
      util.addToHistory(app.optimizeConfigFile)()
      if app.configFile.get("trackingImplementation") == "fastFishTracking.tracking":
        util.addToHistory(adjustFastFishTrackingPage)(useNext=False, wellPositions=wellPositions, wellShape=wellShape, wellIdx=numWell, frameIdx=frameSlider.value())
      else:
        trackingMethod = app.configFile.get("trackingMethod")
        if not trackingMethod:
          if app.configFile.get("headEmbeded", False):
            util.addToHistory(app.calculateBackground)(app, 0, useNext=False, wellPositions=wellPositions, wellShape=wellShape, frameIdx=frameSlider.value())
          else:
            util.addToHistory(app.calculateBackgroundFreelySwim)(app, 0, automaticParameters=True, useNext=False, wellPositions=wellPositions, wellShape=wellShape, wellIdx=numWell, frameIdx=frameSlider.value())
    QTimer.singleShot(0, lambda: adjustButton.clicked.connect(adjustParametrsClicked))
    trackingPointsLayout.addWidget(adjustButton)

  trackingPointsLayout.addStretch(1)
  layout.addLayout(trackingPointsLayout)

  stopTimer = True
  timer = _CustomTimer()
  timer.started.connect(lambda: playBtn.setIcon(pauseIcon))
  timer.stopped.connect(lambda: playBtn.setIcon(playIcon))
  timer.setInterval(1)

  def nextFrame():
    nonlocal stopTimer
    stopTimer = False
    frameSlider.setValue(frameSlider.value() + 1)
    stopTimer = True
  timer.timeout.connect(nextFrame)

  startFrame = getFrame(frameSlider, timer, btnGroup, stopTimer, topLeftCircleCb=topLeftCircleCb)
  timer.start()
  focusWidgets = {frameSlider}
  focusWidgets.add(frameSpinbox)
  QTimer.singleShot(0, lambda: frameSlider.setFocus())
  util.showDialog(layout, title="Video", labelInfo=(startFrame, video), focusWidgets=focusWidgets, exitSignals=(adjustButton.clicked,) if adjustButton is not None else ())
  timer.stop()
  del getFrame



import math

import cv2
import numpy as np


class HeadTrackingHeadingCalculationMixin:
  def _calculateMinDistFromOtherAnimals(self, animal_Id, trackingHeadTailAllAnimals, i):
    mindist   = 1000000
    for animal_Id2 in range(0, self._hyperparameters["nbAnimalsPerWell"]):
      x_curFrame_animal_Id   = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0]
      y_curFrame_animal_Id   = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1]
      if x_curFrame_animal_Id != 0 and y_curFrame_animal_Id != 0:
        if animal_Id != animal_Id2:
          x_curFrame_animal_Id2  = trackingHeadTailAllAnimals[animal_Id2, i-self._firstFrame][0][0]
          y_curFrame_animal_Id2  = trackingHeadTailAllAnimals[animal_Id2, i-self._firstFrame][0][1]
          dist = math.sqrt((x_curFrame_animal_Id-x_curFrame_animal_Id2)**2 + (y_curFrame_animal_Id-y_curFrame_animal_Id2)**2)
          if dist < mindist:
            mindist = dist
    return mindist

  def _multipleAnimalsHeadTrackingAdvance(self, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, gray, i, thresh1, thresh3, lengthX):
    marginX = self._hyperparameters["multipleAnimalTrackingAdvanceAlgorithmMarginX"]

    headCoordinatesOptions    = []
    headCoordinatesOptionsAlt = []
    x = 0
    y = 0
    minAreaCur = self._hyperparameters["minArea"]
    maxAreaCur = self._hyperparameters["maxArea"]

    ret,thresh2 = cv2.threshold(gray,self._hyperparameters["thresholdForBlobImg"],255,cv2.THRESH_BINARY)
    erodeSize = self._hyperparameters["erodeSize"]
    kernel  = np.ones((erodeSize,erodeSize), np.uint8)
    thresh2 = cv2.dilate(thresh2, kernel, iterations=self._hyperparameters["dilateIter"])

    thresh2[:,0] = 255
    thresh2[0,:] = 255
    thresh2[:, len(thresh2[0])-1] = 255
    thresh2[len(thresh2)-1, :]    = 255

    # print("frame:", i)
    # cv2.imshow('Frame', thresh2)
    # cv2.waitKey(0)

    if self._hyperparameters["multipleHeadTrackingIterativelyRelaxAreaCriteria"]:
      contours, hierarchy = cv2.findContours(thresh2,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
      while (len(headCoordinatesOptions) < self._hyperparameters["nbAnimalsPerWell"]) and (minAreaCur > 0):
        # print("MMMinAreaCur:", minAreaCur)
        for contour in contours:
          area = cv2.contourArea(contour)
          # if area > 100:
            # print("area:", area)
          if (area > minAreaCur) and (area < maxAreaCur):
            M = cv2.moments(contour)
            if M['m00']:
              x = int(M['m10']/M['m00'])
              y = int(M['m01']/M['m00'])
            else:
              x = 0
              y = 0
            if not([x, y] in headCoordinatesOptions):
              headCoordinatesOptions.append([x, y])
        minAreaCur = minAreaCur - int(self._hyperparameters["minArea"]/10)
        maxAreaCur = maxAreaCur + int(self._hyperparameters["minArea"]/10)
    else:
      contours, hierarchy = cv2.findContours(thresh2,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
      for contour in contours:
        area = cv2.contourArea(contour)
        if (area > minAreaCur) and (area < maxAreaCur):
          M = cv2.moments(contour)
          if M['m00']:
            x = int(M['m10']/M['m00'])
            y = int(M['m01']/M['m00'])
            # if area < 600:
            headCoordinatesOptions.append([x, y])
            # else:
              # (xE,yE),(MA,ma),angle = cv2.fitEllipse(contour)
              # r = MA * 0.8
              # headCoordinatesOptions.append([int(xE + r*math.cos((math.pi/2)+angle*(math.pi/180))), int(yE + r*math.sin((math.pi/2)+angle*(math.pi/180)))])
              # headCoordinatesOptions.append([int(xE - r*math.cos((math.pi/2)+angle*(math.pi/180))), int(yE - r*math.sin((math.pi/2)+angle*(math.pi/180)))])
          # else:
            # x = 0
            # y = 0
            # headCoordinatesOptions.append([x, y])
      # Alternative head coordinates options
      for contour in contours:
        area = cv2.contourArea(contour)
        if (area > minAreaCur) and (area < maxAreaCur) and (len(contour) > 4):
          M = cv2.moments(contour)
          if M['m00']:
            x = int(M['m10']/M['m00'])
            y = int(M['m01']/M['m00'])
            (xE,yE),(MA,ma),angle = cv2.fitEllipse(contour)
            r = MA * 0.8
            headCoordinatesOptionsAlt.append([int(xE + r*math.cos((math.pi/2)+angle*(math.pi/180))), int(yE + r*math.sin((math.pi/2)+angle*(math.pi/180))), x, y, int(xE - r*math.cos((math.pi/2)+angle*(math.pi/180))), int(yE - r*math.sin((math.pi/2)+angle*(math.pi/180)))])
            headCoordinatesOptionsAlt.append([int(xE - r*math.cos((math.pi/2)+angle*(math.pi/180))), int(yE - r*math.sin((math.pi/2)+angle*(math.pi/180))), x, y, int(xE + r*math.cos((math.pi/2)+angle*(math.pi/180))), int(yE + r*math.sin((math.pi/2)+angle*(math.pi/180)))])
        else:
          if (area > 0) and (area <= minAreaCur):
            M = cv2.moments(contour)
            if M['m00']:
              x = int(M['m10']/M['m00'])
              y = int(M['m01']/M['m00'])
              headCoordinatesOptionsAlt.append([x, y, -1, -1, -1, -1])

    headCoordinatesOptionsAlreadyTakenDist     = [-1 for k in headCoordinatesOptions]
    headCoordinatesOptionsAlreadyTakenAnimalId = [-1 for k in headCoordinatesOptions]
    animalNotPutOrEjectedBecausePositionAlreadyTaken = 1
    if i > self._firstFrame:
      while animalNotPutOrEjectedBecausePositionAlreadyTaken:
        animalNotPutOrEjectedBecausePositionAlreadyTaken = 0
        for animal_Id in range(0, self._hyperparameters["nbAnimalsPerWell"]):
          x_prevFrame_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-1][0][0]
          y_prevFrame_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-1][0][1]
          x_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0]
          y_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1]
          if ((x_prevFrame_animal_Id != 0) and (y_prevFrame_animal_Id != 0) and (x_curFrame_animal_Id == 0) and (y_curFrame_animal_Id == 0)):
            min_dist  = 10000000000000000000000
            index_min = -1
            for idxCoordinateOption, [idx_x_Option, idx_y_Option] in enumerate(headCoordinatesOptions):
              dist = math.sqrt((idx_x_Option - x_prevFrame_animal_Id)**2 + (idx_y_Option - y_prevFrame_animal_Id)**2)
              if dist < min_dist and (dist < self._hyperparameters["multipleAnimalTrackingAdvanceAlgorithmDist1"] or (dist < self._hyperparameters["multipleAnimalTrackingAdvanceAlgorithmDist2"] and idx_x_Option > marginX and idx_x_Option < lengthX - marginX)):
                if headCoordinatesOptionsAlreadyTakenDist[idxCoordinateOption] != -1:
                  if dist < headCoordinatesOptionsAlreadyTakenDist[idxCoordinateOption]:
                    min_dist = dist
                    index_min = idxCoordinateOption
                else:
                  min_dist = dist
                  index_min = idxCoordinateOption
            # if len(headCoordinatesOptions):
            if index_min >= 0:
              if (headCoordinatesOptionsAlreadyTakenDist[index_min] == -1):
                # Save position of animal for frame i-self._firstFrame
                trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] = headCoordinatesOptions[index_min][0]
                trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] = headCoordinatesOptions[index_min][1]
                headCoordinatesOptionsAlreadyTakenDist[index_min]     = min_dist
                headCoordinatesOptionsAlreadyTakenAnimalId[index_min] = animal_Id
              else:
                animalNotPutOrEjectedBecausePositionAlreadyTaken = 1
                if min_dist < headCoordinatesOptionsAlreadyTakenDist[index_min]:
                  # "Ejecting" previously saved position of animal for frame i-self._firstFrame
                  animal_Id_Eject = headCoordinatesOptionsAlreadyTakenAnimalId[index_min]
                  trackingHeadTailAllAnimals[animal_Id_Eject, i-self._firstFrame][0][0] = 0
                  trackingHeadTailAllAnimals[animal_Id_Eject, i-self._firstFrame][0][1] = 0
                  # Replace position of animal for frame i-self._firstFrame
                  trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] = headCoordinatesOptions[index_min][0]
                  trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] = headCoordinatesOptions[index_min][1]
                  headCoordinatesOptionsAlreadyTakenDist[index_min]     = min_dist
                  headCoordinatesOptionsAlreadyTakenAnimalId[index_min] = animal_Id

    headCoordinatesOptions = [headCoordinatesOptions[idx] for idx, k in enumerate(headCoordinatesOptionsAlreadyTakenAnimalId) if k == -1]

    for idxCoordinateOption, [idx_x_Option, idx_y_Option] in enumerate(headCoordinatesOptions):
      for animal_Id in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        if (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] == 0):
          if i == self._firstFrame:
            trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] = idx_x_Option
            trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] = idx_y_Option
            break
          else:
            if (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-1][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-1][0][1] == 0): # NEW !!!!
              trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] = idx_x_Option
              trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] = idx_y_Option
              break

    # Rescuing disparitions in the middle of the thing
    if i > self._firstFrame:
      for animal_Id in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        x_prevFrame_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-1][0][0]
        y_prevFrame_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-1][0][1]
        x_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0]
        y_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1]
        if ((x_prevFrame_animal_Id != 0) and (y_prevFrame_animal_Id != 0) and (x_curFrame_animal_Id == 0) and (y_curFrame_animal_Id == 0) and (x_prevFrame_animal_Id > marginX) and (x_prevFrame_animal_Id < lengthX - marginX) ):
          print("frame:", i, " ; animal:", animal_Id, " has disapeared, looking for alternative")
          min_dist  = 10000000000
          index_min = - 1
          for idxCoordinateOption, [idx_x_Option, idx_y_Option, no1, no1, no3, no4] in enumerate(headCoordinatesOptionsAlt):
            dist = math.sqrt((idx_x_Option - x_prevFrame_animal_Id)**2 + (idx_y_Option - y_prevFrame_animal_Id)**2)
            if dist < min_dist and (dist < self._hyperparameters["multipleAnimalTrackingAdvanceAlgorithmDist1"] or (dist < self._hyperparameters["multipleAnimalTrackingAdvanceAlgorithmDist2"] and idx_x_Option > marginX and idx_x_Option < lengthX - marginX)):
              min_dist = dist
              index_min = idxCoordinateOption
          print("index_min:", index_min)
          mindistFromOtherAnimals = self._calculateMinDistFromOtherAnimals(animal_Id, trackingHeadTailAllAnimals, i)
          if index_min != -1 and mindistFromOtherAnimals >= 5:
            print("coord:", headCoordinatesOptionsAlt[index_min][0], headCoordinatesOptionsAlt[index_min][1], headCoordinatesOptionsAlt[index_min][2], headCoordinatesOptionsAlt[index_min][3], headCoordinatesOptionsAlt[index_min][4], headCoordinatesOptionsAlt[index_min][5])
            trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] = headCoordinatesOptionsAlt[index_min][0]
            trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] = headCoordinatesOptionsAlt[index_min][1]
            # THIS BELLOW IS NEW !!!
            if headCoordinatesOptionsAlt[index_min][2] != -1: # this alt head option is coming from a blob split in half
              for animal_Id2 in range(0, self._hyperparameters["nbAnimalsPerWell"]):
                alt_assigned_x = headCoordinatesOptionsAlt[index_min][0]
                alt_assigned_y = headCoordinatesOptionsAlt[index_min][1]
                alt_center_x   = headCoordinatesOptionsAlt[index_min][2]
                alt_center_y   = headCoordinatesOptionsAlt[index_min][3]
                alt_opo_x      = headCoordinatesOptionsAlt[index_min][4]
                alt_opo_y      = headCoordinatesOptionsAlt[index_min][5]
                if animal_Id2 != animal_Id:
                  otherAnimal_x = trackingHeadTailAllAnimals[animal_Id2, i-self._firstFrame][0][0]
                  otherAnimal_y = trackingHeadTailAllAnimals[animal_Id2, i-self._firstFrame][0][1]
                  if (otherAnimal_x == alt_center_x and otherAnimal_y == alt_center_y) or (otherAnimal_x == alt_assigned_x and otherAnimal_y == alt_assigned_y):
                    print("yeah!!! changed!!", (otherAnimal_x == alt_center_x and otherAnimal_y == alt_center_y), (otherAnimal_x == alt_assigned_x and otherAnimal_y == alt_assigned_y))
                    trackingHeadTailAllAnimals[animal_Id2, i-self._firstFrame][0][0] = alt_opo_x
                    trackingHeadTailAllAnimals[animal_Id2, i-self._firstFrame][0][1] = alt_opo_y
                  break

    # Remove doublons
    if i > self._firstFrame:
      for animal_Id in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        # mindist   = 1000000
        # for animal_Id2 in range(0, self._hyperparameters["nbAnimalsPerWell"]):
          # x_curFrame_animal_Id   = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0]
          # y_curFrame_animal_Id   = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1]
          # if x_curFrame_animal_Id != 0 and y_curFrame_animal_Id != 0:
            # if animal_Id != animal_Id2:
              # x_curFrame_animal_Id2  = trackingHeadTailAllAnimals[animal_Id2, i-self._firstFrame][0][0]
              # y_curFrame_animal_Id2  = trackingHeadTailAllAnimals[animal_Id2, i-self._firstFrame][0][1]
              # dist = math.sqrt((x_curFrame_animal_Id-x_curFrame_animal_Id2)**2 + (y_curFrame_animal_Id-y_curFrame_animal_Id2)**2)
              # if dist < mindist:
                # mindist = dist
        mindist = self._calculateMinDistFromOtherAnimals(animal_Id, trackingHeadTailAllAnimals, i)
        if mindist < 4:
          print("removed ", trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0], trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1])
          trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] = 0
          trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] = 0

    count = 0
    if i > self._firstFrame:
      for animal_Id in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        x_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0]
        y_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1]
        if x_curFrame_animal_Id != 0 and y_curFrame_animal_Id != 0:
          count = count + 1
    print("Frame ", i, "; count: ", count)

    # for idxCoordinateOption, [idx_x_Option, idx_y_Option] in enumerate(headCoordinatesOptions):
      # print(i, idxCoordinateOption, idx_x_Option, idx_y_Option)
      # min_dist  = 10000000000000000000000
      # index_min_animal_Id = -1
      # maxdepth_min_animal_Id = 0
      # index_animal_Id_alwaysBeenAt0 = -1
      # for animal_Id in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        # if (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] == 0):
          # depth = 1
          # while (i - self._firstFrame - depth >= 0) and (depth < 20) and (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-depth][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-depth][0][1] == 0):
            # depth = depth + 1
          # xPast_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-depth][0][0]
          # yPast_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-depth][0][1]
          # if (xPast_animal_Id != 0) or (yPast_animal_Id != 0):
            # print("animal_Id:", animal_Id, "; (xPast_animal_Id, yPast_animal_Id):", xPast_animal_Id, yPast_animal_Id)
            # dist = math.sqrt((idx_x_Option - xPast_animal_Id)**2 + (idx_y_Option - yPast_animal_Id)**2)
            # print("dist:", dist)
            # if dist < 30:
              # print("depth", depth)
              # if dist < min_dist:
                # min_dist = dist
                # index_min_animal_Id = animal_Id
                # maxdepth_min_animal_Id = depth
          # else:
            # if index_animal_Id_alwaysBeenAt0 == -1:
              # index_animal_Id_alwaysBeenAt0 = animal_Id
      # print("index_min_animal_Id:", index_min_animal_Id)
      # if index_min_animal_Id >= 0:
        # trackingHeadTailAllAnimals[index_min_animal_Id, i-self._firstFrame][0][0] = idx_x_Option
        # trackingHeadTailAllAnimals[index_min_animal_Id, i-self._firstFrame][0][1] = idx_y_Option
        # x_maxPast_animal_Id = trackingHeadTailAllAnimals[index_min_animal_Id, i-self._firstFrame-maxdepth_min_animal_Id][0][0]
        # y_maxPast_animal_Id = trackingHeadTailAllAnimals[index_min_animal_Id, i-self._firstFrame-maxdepth_min_animal_Id][0][1]
        # print("x_maxPast_animal_Id, y_maxPast_animal_Id:", x_maxPast_animal_Id, y_maxPast_animal_Id)
        # print("maxdepth_min_animal_Id", maxdepth_min_animal_Id)
        # x_step = (idx_x_Option - x_maxPast_animal_Id) / maxdepth_min_animal_Id
        # y_step = (idx_y_Option - y_maxPast_animal_Id) / maxdepth_min_animal_Id
        # for blankSpace in range(1, maxdepth_min_animal_Id):
          # trackingHeadTailAllAnimals[index_min_animal_Id, i-self._firstFrame-maxdepth_min_animal_Id+blankSpace][0][0] = x_maxPast_animal_Id + blankSpace * x_step
          # trackingHeadTailAllAnimals[index_min_animal_Id, i-self._firstFrame-maxdepth_min_animal_Id+blankSpace][0][1] = y_maxPast_animal_Id + blankSpace * y_step
      # else:
        # if index_animal_Id_alwaysBeenAt0 != -1:
          # trackingHeadTailAllAnimals[index_animal_Id_alwaysBeenAt0, i-self._firstFrame][0][0] = idx_x_Option
          # trackingHeadTailAllAnimals[index_animal_Id_alwaysBeenAt0, i-self._firstFrame][0][1] = idx_y_Option

    for animal_Id in range(0, self._hyperparameters["nbAnimalsPerWell"]):
      x_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0]
      y_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1]
      if x_curFrame_animal_Id != 0 and y_curFrame_animal_Id != 0:
        x_curFrame_animal_Id = int(x_curFrame_animal_Id)
        y_curFrame_animal_Id = int(y_curFrame_animal_Id)
        heading = 0 #calculateHeadingSimple(x_curFrame_animal_Id, y_curFrame_animal_Id, thresh2, self._hyperparameters)
        trackingHeadingAllAnimals[animal_Id, i-self._firstFrame] = heading

  @staticmethod
  def __simpleOptimalValueSearch(PtClosest, contour, unitVector, lenX, lenY):

    factor = 0
    dist = 1
    maxDist = 0
    indMax = 0
    testCenter = PtClosest + factor * unitVector
    while (dist > 0) and (factor < 20) and (testCenter[0] >= 0) and (testCenter[1] >= 0) and (testCenter[0] < lenX) and (testCenter[1] < lenY):
      factor = factor + 1
      testCenter = PtClosest + factor * unitVector
      testCenter = testCenter.astype(int)
      dist = cv2.pointPolygonTest(contour, (float(testCenter[0]), float(testCenter[1])), True)
      if dist > maxDist:
        maxDist = dist
        indMax  = factor

    testCenter = PtClosest + indMax * unitVector
    testCenter = testCenter.astype(int)

    return testCenter

  def __reajustCenterOfMassIfNecessary(self, contour, x, y, lenX, lenY):
    inside = cv2.pointPolygonTest(contour, (float(x), float(y)), True)
    if inside < 0:

      minDist = 100000000000000
      indMin  = 0
      for i in range(0, len(contour)):
        Pt = contour[i][0]
        dist = math.sqrt((Pt[0] - x)**2 + (Pt[1] - y)**2)
        if dist < minDist:
          minDist = dist
          indMin  = i
      PtClosest = contour[indMin][0]
      unitVector = np.array([PtClosest[0] - x, PtClosest[1] - y])
      unitVectorLength = math.sqrt(unitVector[0]**2 + unitVector[1]**2)
      unitVector[0] = unitVector[0] / unitVectorLength
      unitVector[1] = unitVector[1] / unitVectorLength
      if True:
        factor = 5
        testCenter = PtClosest + factor * unitVector
        testCenter = testCenter.astype(int)
        while (cv2.pointPolygonTest(contour, (float(testCenter[0]), float(testCenter[1])), True) <= 0) and (factor > 1):
          factor = factor - 1
          testCenter = PtClosest + factor * unitVector
      else:
        testCenter = self.__simpleOptimalValueSearch(PtClosest, contour, unitVector, lenX, lenY)

      x = testCenter[0]
      y = testCenter[1]

    return [x, y]

  @staticmethod
  def __findCenterByIterativelyDilating(initialContour, lenX, lenY):
    x = 0
    y = 0
    image = np.zeros((lenY, lenX))
    image[:, :] = 255
    image = image.astype(np.uint8)
    kernel  = np.ones((3, 3), np.uint8)
    if type(initialContour) != int:
      cv2.fillPoly(image, pts =[initialContour], color=(0))
      image[:,0] = 255
      image[0,:] = 255
      image[:, len(image[0])-1] = 255
      image[len(image)-1, :]    = 255
      nbBlackPixels = 1
      dilateIter = 0
      while nbBlackPixels > 0:
        dilateIter   = dilateIter + 1
        dilatedImage = cv2.dilate(image, kernel, iterations=dilateIter)
        nbBlackPixels = cv2.countNonZero(255-dilatedImage)
      dilateIter   = dilateIter - 1
      dilatedImage = cv2.dilate(image, kernel, iterations=dilateIter)
      dilatedImage[:,0] = 255
      dilatedImage[0,:] = 255
      dilatedImage[:, len(dilatedImage[0])-1] = 255
      dilatedImage[len(dilatedImage)-1, :]    = 255
      contours, hierarchy = cv2.findContours(dilatedImage, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
      maxContour = 0
      maxContourArea = 0
      for contour in contours:
        contourArea = cv2.contourArea(contour)
        if contourArea < int(lenX * lenY * 0.8):
          if contourArea > maxContourArea:
            maxContourArea = contourArea
            maxContour     = contour
      M = cv2.moments(maxContour)
      if M['m00']:
        x = int(M['m10']/M['m00'])
        y = int(M['m01']/M['m00'])
    return [x, y]

  def _multipleAnimalsHeadTracking(self, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, gray, i, thresh1, xmin=0, ymin=0):

    headCoordinatesOptions      = []
    headCoordinatesAreasOptions = []
    x = 0
    y = 0
    minAreaCur = self._hyperparameters["minArea"]
    maxAreaCur = self._hyperparameters["maxArea"]
    meanArea   = (self._hyperparameters["minArea"] + self._hyperparameters["maxArea"]) / 2

    thresh2 = thresh1.copy()
    erodeSize = self._hyperparameters["erodeSize"]
    if erodeSize:
      kernel  = np.ones((erodeSize,erodeSize), np.uint8)
      thresh2 = cv2.dilate(thresh2, kernel, iterations=self._hyperparameters["dilateIter"])

    thresh2[:,0] = 255
    thresh2[0,:] = 255
    thresh2[:, len(thresh2[0])-1] = 255
    thresh2[len(thresh2)-1, :]    = 255

    if self._hyperparameters["multipleHeadTrackingIterativelyRelaxAreaCriteria"]:
      contours, hierarchy = cv2.findContours(thresh2,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
      areaDecreaseFactor = int(self._hyperparameters["minArea"]/10)
      while (len(headCoordinatesOptions) < self._hyperparameters["nbAnimalsPerWell"]) and (minAreaCur > 0):
        headCoordinatesOptions = []
        headCoordinatesAreasOptions = []
        for contour in contours:
          area = cv2.contourArea(contour)
          if (area > minAreaCur) and (area < maxAreaCur):
            if self._hyperparameters["findCenterOfAnimalByIterativelyDilating"]:
              [x, y] = self.__findCenterByIterativelyDilating(contour, len(thresh2[0]), len(thresh2))
            else:
              M = cv2.moments(contour)
              if M['m00']:
                x = int(M['m10']/M['m00'])
                y = int(M['m01']/M['m00'])
              else:
                x = 0
                y = 0
              if self._hyperparameters["readjustCenterOfMassIfNotInsideContour"]:
                [x, y] = self.__reajustCenterOfMassIfNecessary(contour, x, y, len(thresh2[0]), len(thresh2))
            if not([x, y] in headCoordinatesOptions):
              headCoordinatesOptions.append([x+xmin, y+ymin])
              headCoordinatesAreasOptions.append(abs(area - (meanArea/2)))

        if areaDecreaseFactor:
          minAreaCur = minAreaCur - areaDecreaseFactor
          maxAreaCur = int(maxAreaCur * 1.1)
        else:
          minAreaCur = minAreaCur - 1
          maxAreaCur = int(maxAreaCur * 1.1)
    else:
      contours, hierarchy = cv2.findContours(thresh2,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
      for contour in contours:
        area = cv2.contourArea(contour)
        if (area > minAreaCur) and (area < maxAreaCur):
          # print("Find center of mass: area:", area)
          if self._hyperparameters["findCenterOfAnimalByIterativelyDilating"]:
            [x, y] = self.__findCenterByIterativelyDilating(contour, len(thresh2[0]), len(thresh2))
          else:
            M = cv2.moments(contour)
            if M['m00']:
              x = int(M['m10']/M['m00'])
              y = int(M['m01']/M['m00'])
            else:
              x = 0
              y = 0
            if self._hyperparameters["readjustCenterOfMassIfNotInsideContour"]:
              [x, y] = self.__reajustCenterOfMassIfNecessary(contour, x, y, len(thresh2[0]), len(thresh2))
          headCoordinatesOptions.append([x+xmin, y+ymin])
          headCoordinatesAreasOptions.append(abs(area - (meanArea/2)))

    headCoordinatesAreasOptionsSorted = np.argsort(headCoordinatesAreasOptions)
    headCoordinatesOptions = [headCoordinatesOptions[elem] for elem in headCoordinatesAreasOptionsSorted]

    headCoordinatesOptionsAlreadyTakenDist     = [-1 for k in headCoordinatesOptions]
    headCoordinatesOptionsAlreadyTakenAnimalId = [-1 for k in headCoordinatesOptions]
    animalNotPutOrEjectedBecausePositionAlreadyTaken = 1
    if i > self._firstFrame:
      while animalNotPutOrEjectedBecausePositionAlreadyTaken:
        animalNotPutOrEjectedBecausePositionAlreadyTaken = 0
        for animal_Id in range(0, self._hyperparameters["nbAnimalsPerWell"]):
          x_prevFrame_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-1][0][0]
          y_prevFrame_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-1][0][1]
          x_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0]
          y_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1]
          if ((x_prevFrame_animal_Id != 0) and (y_prevFrame_animal_Id != 0) and (x_curFrame_animal_Id == 0) and (y_curFrame_animal_Id == 0)):
            min_dist  = 10000000000000000000000
            index_min = -1
            for idxCoordinateOption, [idx_x_Option, idx_y_Option] in enumerate(headCoordinatesOptions):
              dist = (idx_x_Option - x_prevFrame_animal_Id)**2 + (idx_y_Option - y_prevFrame_animal_Id)**2
              if dist < min_dist:
                if headCoordinatesOptionsAlreadyTakenDist[idxCoordinateOption] != -1:
                  if dist < headCoordinatesOptionsAlreadyTakenDist[idxCoordinateOption]:
                    min_dist = dist
                    index_min = idxCoordinateOption
                else:
                  min_dist = dist
                  index_min = idxCoordinateOption
            # if len(headCoordinatesOptions):
            if index_min >= 0:
              if (headCoordinatesOptionsAlreadyTakenDist[index_min] == -1):
                # Save position of animal for frame i-self._firstFrame
                trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] = headCoordinatesOptions[index_min][0]
                trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] = headCoordinatesOptions[index_min][1]
                headCoordinatesOptionsAlreadyTakenDist[index_min]     = min_dist
                headCoordinatesOptionsAlreadyTakenAnimalId[index_min] = animal_Id
              else:
                animalNotPutOrEjectedBecausePositionAlreadyTaken = 1
                if min_dist < headCoordinatesOptionsAlreadyTakenDist[index_min]:
                  # "Ejecting" previously saved position of animal for frame i-self._firstFrame
                  animal_Id_Eject = headCoordinatesOptionsAlreadyTakenAnimalId[index_min]
                  trackingHeadTailAllAnimals[animal_Id_Eject, i-self._firstFrame][0][0] = 0
                  trackingHeadTailAllAnimals[animal_Id_Eject, i-self._firstFrame][0][1] = 0
                  # Replace position of animal for frame i-self._firstFrame
                  trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] = headCoordinatesOptions[index_min][0]
                  trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] = headCoordinatesOptions[index_min][1]
                  headCoordinatesOptionsAlreadyTakenDist[index_min]     = min_dist
                  headCoordinatesOptionsAlreadyTakenAnimalId[index_min] = animal_Id

    headCoordinatesOptions = [headCoordinatesOptions[idx] for idx, k in enumerate(headCoordinatesOptionsAlreadyTakenAnimalId) if k == -1]

    for idxCoordinateOption, [idx_x_Option, idx_y_Option] in enumerate(headCoordinatesOptions):
      for animal_Id in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        if (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] == 0):
          if i == self._firstFrame:
            trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] = idx_x_Option
            trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] = idx_y_Option
            break
          else:
            if (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-1][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-1][0][1] == 0): # NEW !!!!
              trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] = idx_x_Option
              trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] = idx_y_Option
              break

    if self._hyperparameters["postProcessMultipleTrajectories"]:
      maxDistanceAuthorized = self._hyperparameters["postProcessMaxDistanceAuthorized"]
      maxDisapearanceFrames = self._hyperparameters["postProcessMaxDisapearanceFrames"]
      for animal_Id in range(0, len(trackingHeadTailAllAnimals)):
        goBackFrames = 1
        xCur = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0]
        yCur = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1]
        if xCur != 0 or yCur != 0:
          while (i-self._firstFrame-goBackFrames >= 0) and (goBackFrames < maxDisapearanceFrames) and (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-goBackFrames][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-goBackFrames][0][1] == 0):
            goBackFrames = goBackFrames + 1
          if (goBackFrames < maxDisapearanceFrames) and (i-self._firstFrame-goBackFrames >= 0):
            xBef = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-goBackFrames][0][0]
            yBef = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-goBackFrames][0][1]
            if math.sqrt((xCur - xBef)**2 + (yCur - yBef)**2) > maxDistanceAuthorized:
              trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] = 0
              trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] = 0

    # for idxCoordinateOption, [idx_x_Option, idx_y_Option] in enumerate(headCoordinatesOptions):
      # print(i, idxCoordinateOption, idx_x_Option, idx_y_Option)
      # min_dist  = 10000000000000000000000
      # index_min_animal_Id = -1
      # maxdepth_min_animal_Id = 0
      # index_animal_Id_alwaysBeenAt0 = -1
      # for animal_Id in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        # if (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1] == 0):
          # depth = 1
          # while (i - self._firstFrame - depth >= 0) and (depth < 20) and (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-depth][0][0] == 0) and (trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-depth][0][1] == 0):
            # depth = depth + 1
          # xPast_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-depth][0][0]
          # yPast_animal_Id = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame-depth][0][1]
          # if (xPast_animal_Id != 0) or (yPast_animal_Id != 0):
            # print("animal_Id:", animal_Id, "; (xPast_animal_Id, yPast_animal_Id):", xPast_animal_Id, yPast_animal_Id)
            # dist = math.sqrt((idx_x_Option - xPast_animal_Id)**2 + (idx_y_Option - yPast_animal_Id)**2)
            # print("dist:", dist)
            # if dist < 30:
              # print("depth", depth)
              # if dist < min_dist:
                # min_dist = dist
                # index_min_animal_Id = animal_Id
                # maxdepth_min_animal_Id = depth
          # else:
            # if index_animal_Id_alwaysBeenAt0 == -1:
              # index_animal_Id_alwaysBeenAt0 = animal_Id
      # print("index_min_animal_Id:", index_min_animal_Id)
      # if index_min_animal_Id >= 0:
        # trackingHeadTailAllAnimals[index_min_animal_Id, i-self._firstFrame][0][0] = idx_x_Option
        # trackingHeadTailAllAnimals[index_min_animal_Id, i-self._firstFrame][0][1] = idx_y_Option
        # x_maxPast_animal_Id = trackingHeadTailAllAnimals[index_min_animal_Id, i-self._firstFrame-maxdepth_min_animal_Id][0][0]
        # y_maxPast_animal_Id = trackingHeadTailAllAnimals[index_min_animal_Id, i-self._firstFrame-maxdepth_min_animal_Id][0][1]
        # print("x_maxPast_animal_Id, y_maxPast_animal_Id:", x_maxPast_animal_Id, y_maxPast_animal_Id)
        # print("maxdepth_min_animal_Id", maxdepth_min_animal_Id)
        # x_step = (idx_x_Option - x_maxPast_animal_Id) / maxdepth_min_animal_Id
        # y_step = (idx_y_Option - y_maxPast_animal_Id) / maxdepth_min_animal_Id
        # for blankSpace in range(1, maxdepth_min_animal_Id):
          # trackingHeadTailAllAnimals[index_min_animal_Id, i-self._firstFrame-maxdepth_min_animal_Id+blankSpace][0][0] = x_maxPast_animal_Id + blankSpace * x_step
          # trackingHeadTailAllAnimals[index_min_animal_Id, i-self._firstFrame-maxdepth_min_animal_Id+blankSpace][0][1] = y_maxPast_animal_Id + blankSpace * y_step
      # else:
        # if index_animal_Id_alwaysBeenAt0 != -1:
          # trackingHeadTailAllAnimals[index_animal_Id_alwaysBeenAt0, i-self._firstFrame][0][0] = idx_x_Option
          # trackingHeadTailAllAnimals[index_animal_Id_alwaysBeenAt0, i-self._firstFrame][0][1] = idx_y_Option

    if self._hyperparameters["findCenterOfAnimalByIterativelyDilating"] == 0 or self._hyperparameters["trackTail"] == 0:
      for animal_Id in range(0, self._hyperparameters["nbAnimalsPerWell"]):
        x_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][0]
        y_curFrame_animal_Id  = trackingHeadTailAllAnimals[animal_Id, i-self._firstFrame][0][1]
        if x_curFrame_animal_Id != 0 and y_curFrame_animal_Id != 0:
          x_curFrame_animal_Id = int(x_curFrame_animal_Id)
          y_curFrame_animal_Id = int(y_curFrame_animal_Id)

          # Removing the other blobs from the image to get good heading detection
          # Performance improvement possible below: TODO in the future
          correspondingContour = 0
          contours, hierarchy = cv2.findContours(thresh2,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
          for contour in contours:
            dist = cv2.pointPolygonTest(contour, (float(x_curFrame_animal_Id), float(y_curFrame_animal_Id)), True)
            if dist >= 0:
              correspondingContour = contour
          thresh2bis = np.zeros((len(thresh2), len(thresh2[0])))
          thresh2bis[:, :] = 255
          if type(correspondingContour) != int:
            cv2.fillPoly(thresh2bis, pts =[correspondingContour], color=(0, 0, 0))

          if self._hyperparameters["forceBlobMethodForHeadTracking"]: # Fish tracking, usually
            previousFrameHeading = trackingHeadingAllAnimals[animal_Id, i-self._firstFrame-1] if i-self._firstFrame-1 > 0 else 0
            [heading, lastFirstTheta] = self._calculateHeading(x_curFrame_animal_Id, y_curFrame_animal_Id, 0, thresh1, thresh2bis, 0, previousFrameHeading)
          else: # Drosophilia, mouse, and other center of mass only tracking
            heading = self._calculateHeadingSimple(x_curFrame_animal_Id, y_curFrame_animal_Id, thresh2bis)

          trackingHeadingAllAnimals[animal_Id, i-self._firstFrame] = heading

  def __headTrackingTakeHeadClosestToWellCenter(self, thresh1, thresh2, blur, erodeSize, minArea, maxArea, frame_width, frame_height):
    x = 0
    y = 0
    xNew = 0
    yNew = 0
    minAreaCur = minArea
    maxAreaCur = maxArea
    while (x == 0) and (minAreaCur > -200):
      contours, hierarchy = cv2.findContours(thresh2,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
      for contour in contours:
        area = cv2.contourArea(contour)
        if (area > minAreaCur) and (area < maxAreaCur):
          M = cv2.moments(contour)
          xNew = int(M['m10']/M['m00'])
          yNew = int(M['m01']/M['m00'])
          distToCenterNew = math.sqrt((xNew-200)**2 + (yNew-200)**2)
          distToCenter    = math.sqrt((x-200)**2    + (y-200)**2)
          # print("distToCenter:",distToCenter," ; distToCenterNew:",distToCenterNew)
          if distToCenterNew < distToCenter:
            x = xNew
            y = yNew
            # print("change realized, new center is:", x, y)
      minAreaCur = minAreaCur - 100
      maxAreaCur = maxAreaCur + 100

    halfDiam = 50
    xmin = self._assignValueIfBetweenRange(x - halfDiam, 0, frame_width-1)
    xmax = self._assignValueIfBetweenRange(x + halfDiam, 0, frame_height-1)
    ymin = self._assignValueIfBetweenRange(y - halfDiam, 0, frame_width-1)
    ymax = self._assignValueIfBetweenRange(y + halfDiam, 0, frame_height-1)
    blur[0:ymin, :]              = 255
    blur[ymax:frame_height-1, :] = 255
    blur[:, 0:xmin]              = 255
    blur[:, xmax:frame_width-1]  = 255
    (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(blur)

    return headPosition

  def _headTrackingHeadingCalculation(self, i, blur, thresh1, thresh2, gray, erodeSize, frame_width, frame_height, trackingHeadingAllAnimals, trackingHeadTailAllAnimals, trackingProbabilityOfGoodDetection, headPosition, lengthX, xmin=0, ymin=0, wellNumber=-1):
    xHB_TN = 0
    heading = 0
    x = 0
    y = 0
    lastFirstTheta = 0

    if self._hyperparameters["fixedHeadPositionX"] != -1:

      trackingHeadTailAllAnimals[0, i-self._firstFrame][0][0] = int(self._hyperparameters["fixedHeadPositionX"])
      trackingHeadTailAllAnimals[0, i-self._firstFrame][0][1] = int(self._hyperparameters["fixedHeadPositionY"])

    else:

      if self._hyperparameters["nbAnimalsPerWell"] > 1 or self._hyperparameters["forceBlobMethodForHeadTracking"]:

        if self._hyperparameters["multipleAnimalTrackingAdvanceAlgorithm"] == 0:
          self._multipleAnimalsHeadTracking(trackingHeadingAllAnimals, trackingHeadTailAllAnimals, gray, i, thresh1, xmin, ymin)
        else:
          self._multipleAnimalsHeadTrackingAdvance(trackingHeadingAllAnimals, trackingHeadTailAllAnimals, gray, i, thresh1, thresh2, lengthX)

        heading = 0
        headPosition = [0, 0]
        x = 0
        y = 0
        lastFirstTheta = 0

      else:

        if (self._hyperparameters["headEmbeded"] == 1 and i == self._firstFrame) or (self._hyperparameters["headEmbeded"] == 0) or (self._hyperparameters["headEmbededTeresaNicolson"] == 1):

          if self._hyperparameters["findHeadPositionByUserInput"] == 0:

            if type(blur) == int: # it won't be equal to int for images coming from the faster screen 'multiprocessing'
              paramGaussianBlur = int((self._hyperparameters["paramGaussianBlur"] / 2)) * 2 + 1
              blur = cv2.GaussianBlur(gray, (paramGaussianBlur, paramGaussianBlur), 0)

            # Finds head position for frame i
            takeTheHeadClosestToTheCenter = self._hyperparameters["takeTheHeadClosestToTheCenter"]
            if takeTheHeadClosestToTheCenter == 0:
              (minVal, maxVal, headPosition, maxLoc) = cv2.minMaxLoc(blur)
              if type(trackingProbabilityOfGoodDetection) != int and len(trackingProbabilityOfGoodDetection) and i-self._firstFrame < len(trackingProbabilityOfGoodDetection[0]):
                trackingProbabilityOfGoodDetection[0, i-self._firstFrame] = np.sum(255 - blur)
            else:
              headPosition = self.__headTrackingTakeHeadClosestToWellCenter(thresh1, thresh2, blur, erodeSize, self._hyperparameters["minArea"], self._hyperparameters["maxArea"], frame_width, frame_height)

            x = headPosition[0]
            y = headPosition[1]

            if (self._hyperparameters["headEmbededTeresaNicolson"] == 1) and (i == self._firstFrame):
              xHB_TN = x

            if (self._hyperparameters["headEmbededTeresaNicolson"] == 1):
              headPosition = [xHB_TN + 100, y]

            if type(headPosition) == tuple:
              headPosition = list(headPosition)
              headPosition[0] = headPosition[0] + xmin
              headPosition[1] = headPosition[1] + ymin
              headPosition = tuple(headPosition)
            else:
              headPosition[0] = headPosition[0] + xmin
              headPosition[1] = headPosition[1] + ymin

            # Calculate heading for frame i
            if type(thresh1) != int:
              [heading, lastFirstTheta] = self._calculateHeading(x, y, i, thresh1, thresh2, takeTheHeadClosestToTheCenter, 0, wellNumber)

            if (self._hyperparameters["headEmbededTeresaNicolson"] == 1):
              heading = 0
              lastFirstTheta = 0

            trackingHeadTailAllAnimals[0, i-self._firstFrame][0][0] = headPosition[0]
            trackingHeadTailAllAnimals[0, i-self._firstFrame][0][1] = headPosition[1]
            trackingHeadingAllAnimals[0, i-self._firstFrame] = heading

          else:

            # This is for the head-embedeed: at this point, this is set again in the tail tracking (the heading is set in the tail tracking as well)
            trackingHeadTailAllAnimals[0, i-self._firstFrame][0][0] = headPosition[0]
            trackingHeadTailAllAnimals[0, i-self._firstFrame][0][1] = headPosition[1]

        else:
          # If head embeded, heading and head position stay the same for all frames
          trackingHeadingAllAnimals[0, i-self._firstFrame]  = trackingHeadingAllAnimals[0, 0]
          trackingHeadTailAllAnimals[0, i-self._firstFrame] = trackingHeadTailAllAnimals[0, 0]

    return lastFirstTheta



from ._openResultsFile import openResultsFile


def getFPSandPixelSize(videoName: str) -> [float, float]:
  
  with openResultsFile(videoName, 'r') as results:
    videoFPS = results.attrs['videoFPS']
    videoPixelSize = results.attrs['videoPixelSize']
  
  return [videoFPS, videoPixelSize]



import math

import cv2
import numpy as np
from numpy import linspace
from scipy.interpolate import interp1d
from scipy.interpolate import splprep, splev
from scipy.optimize import curve_fit

from ._tailTrackingBase import TailTrackingBase

applyDilateErode  = True
printDebbbbug     = False
maxThresholdMinus = 20
#

class HeadEmbeddedTailTrackingForImageMixin(TailTrackingBase):

  def __smoothTail(self, points, nbTailPoints, smoothingFactor):
    points3 = [np.array(points[0]), np.array(points[1])]

    y = points[0]
    x = linspace(0, 1, len(y))

    if len(x) > 3:
      interpolated_points = {}

      try:
        if smoothingFactor != -1:
          tck, u = splprep(points3, s=smoothingFactor)
        else:
          tck, u = splprep(points3)
        u2 = [i/(nbTailPoints-1) for i in range(0, nbTailPoints)]
        interpolated_points = splev(u2, tck)
        newX = interpolated_points[0]
        newY = interpolated_points[1]
      except:
        newX = points[0]
        newY = points[1]

    else:
      newX = points[0]
      newY = points[1]

    return [newX, newY]

  def __findNextPoints(self, depth, x, y, frame, points, angle, maxDepth, steps, nbList, initialImage, debug, dontChooseThisPoint = [], maxRadiusForDontChoosePoint = 0):
    lenX = len(frame[0]) - 1
    lenY = len(frame) - 1

    thetaDiffAccept = 1

    if depth < self._hyperparameters["initialTailPortionMaxSegmentDiffAngleCutOffPos"] * maxDepth:
      thetaDiffAccept = self._hyperparameters["initialTailPortionMaxSegmentDiffAngleValue"]

    if depth > 0.85*maxDepth:
      thetaDiffAccept = 0.6

    if self._hyperparameters["headEmbededMaxAngleBetweenSubsequentSegments"]:
      thetaDiffAccept = self._hyperparameters["headEmbededMaxAngleBetweenSubsequentSegments"]

    pixTotMax = 1000000
    maxTheta  = angle

    l = [i*(math.pi/nbList) for i in range(0,2*nbList) if self._distBetweenThetas(i*(math.pi/nbList), angle) < thetaDiffAccept]

    # if debug:
      # print("debug")

    for step in steps:

      if (step < maxDepth - depth) or (step == steps[0]):

        for theta in l:

          xNew = self._assignValueIfBetweenRange(int(x + step * (math.cos(theta))), 0, lenX)
          yNew = self._assignValueIfBetweenRange(int(y + step * (math.sin(theta))), 0, lenY)
          pixTot = frame[yNew][xNew]

          # if debug:
            # print([theta,pixTot])

          # Keeps that theta angle as maximum if appropriate
          if (pixTot < pixTotMax):
            if (len(dontChooseThisPoint) == 0):
              pixTotMax = pixTot
              maxTheta = theta
              xTot = xNew
              yTot = yNew
            else:
              dist = 1000000000000000000
              for num in range(0, len(dontChooseThisPoint[0])):
                dist = min(dist, math.sqrt((xNew - dontChooseThisPoint[0, num])**2 + (yNew - dontChooseThisPoint[1, num])**2))
              if (not(dist <= maxRadiusForDontChoosePoint)):
                pixTotMax = pixTot
                maxTheta = theta
                xTot = xNew
                yTot = yNew

    w = 4
    ym = yTot - w
    yM = yTot + w
    xm = xTot - w
    xM = xTot + w
    if ym < 0:
      ym = 0
    if xm < 0:
      xm = 0
    if yM > len(initialImage):
      yM = len(initialImage)
    if xM > len(initialImage[0]):
      xM = len(initialImage[0])

    pixSur = np.min(frame[ym:yM, xm:xM]) #initialImage[ym:yM, xm:xM])
    # if debug:
      # print("depth:", depth, " ; maxDepth:", maxDepth, " ; pixSur:", pixSur)

    # if depth > 0.95*maxDepth:
      # pixTot = frame[y][x]
      # if (pixTot < pixTotMax):
        # pixTotMax = pixTot
        # maxTheta = theta
        # xTot = x
        # yTot = y
        # depth = maxDepth + 10

    # if debug:
      # print(["max:",maxTheta,pixTotMax])

    # Calculates distance between new and old point
    distSubsquentPoints = math.sqrt((xTot - x)**2 + (yTot - y)**2)

    pixSurMax = self._hyperparameters["headEmbededParamTailDescentPixThreshStop"]
    # pixSurMax = 220 #150 #245 #150
    if depth + distSubsquentPoints < maxDepth and ((pixSur < pixSurMax) or (depth < self._hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth)):
      points = self._appendPoint(xTot, yTot, points)
    else:
      vectX = xTot - x
      vectY = yTot - y
      xTot  = int(x + (maxDepth / (depth + distSubsquentPoints)) * vectX)
      yTot  = int(y + (maxDepth / (depth + distSubsquentPoints)) * vectY)
      points = self._appendPoint(xTot, yTot, points)
    if debug:
      cv2.circle(frame, (xTot, yTot), 3, (255,0,0),   -1)
      self._debugFrame(frame, title='HeadEmbeddedTailTracking')

    newTheta = self._calculateAngle(x,y,xTot,yTot)
    if distSubsquentPoints > 0 and depth + distSubsquentPoints < maxDepth and ((pixSur < pixSurMax) or (depth < self._hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth)):
      (points,nop) = self.__findNextPoints(depth+distSubsquentPoints,xTot,yTot,frame,points,newTheta,maxDepth,steps,nbList,initialImage,debug)

    if depth == 0:
      lenPoints = len(points[0]) - 1
      if points[0, lenPoints-1] == points[0, lenPoints] and points[1, lenPoints-1] == points[1, lenPoints]:
        points = points[:, :len(points[0])-1]

    return (points,newTheta)

  def _headEmbededTailTrackingForImage(self, headPosition, i, frame, maxDepth, tailTip, trackingHeadingAllAnimals=[]):
    steps   = self._hyperparameters["step"]
    nbList  = 10 if self._hyperparameters["nbList"] == -1 else self._hyperparameters["nbList"]

    x = headPosition[0]
    y = headPosition[1]

    initialImage = frame.copy()
    
    if printDebbbbug:
      self._debugFrame(frame, title='Test1')
    
    back = frame.copy()
    back[:, :] = np.median(back[:3, :], axis=0)
    putToWhite = ( frame.astype('int32') >= (back.astype('int32') - 0) )
    frame[putToWhite] = 255
    
    ret, threshEyes = cv2.threshold(frame.copy(), 25, 255, cv2.THRESH_BINARY)
    
    
    ret, frame = cv2.threshold(frame.copy(), 254, 255, cv2.THRESH_BINARY)
    
    if applyDilateErode:
      erodeDilateSize = 25
      frame = cv2.erode(frame, np.ones((erodeDilateSize, erodeDilateSize), np.uint8), iterations=1)
      frame = cv2.dilate(frame, np.ones((erodeDilateSize, erodeDilateSize), np.uint8), iterations=1)
    
    # homogeneous = frame.copy()
    # homogeneous[:, :] = np.median(frame) - maxThresholdMinus #maxThreshold
    
    # frame = np.maximum(frame, homogeneous)
    
    if printDebbbbug:
      self._debugFrame(frame, title='Test2')

    gaussian_blur = self._hyperparameters["headEmbededParamGaussianBlur"]

    frame = cv2.GaussianBlur(frame, (gaussian_blur, gaussian_blur), 0)
    
    if printDebbbbug:
      self._debugFrame(frame, title='Test3')

    if ("initialTailForMaxDepthCalculNotStraight" in self._hyperparameters) and self._hyperparameters["initialTailForMaxDepthCalculNotStraight"]:
      angle = self._hyperparameters["headEmbededParamInitialAngle"]
    else:
      angle = self._calculateAngle(x, y, tailTip[0], tailTip[1])

    points = np.zeros((2, 0))

    (points, lastFirstTheta2) = self.__findNextPoints(0,x,y,frame,points,angle,maxDepth,steps,nbList,initialImage,self._hyperparameters["debugHeadEmbededFindNextPoints"])
    points = np.insert(points, 0, headPosition, axis=1)
    # points = np.insert(points, 0, [cX, cY], axis=1)

    # Anomalie detection here
    headEmbededRetrackIfWeirdInitialTracking = self._hyperparameters["headEmbededRetrackIfWeirdInitialTracking"]
    if headEmbededRetrackIfWeirdInitialTracking:
      points = self.__retrackIfWeirdInitialTracking(points, headPosition, tailTip, x, y, frame, angle, maxDepth, nbList, initialImage, i)

    if len(points[0]) > 3:
      if self._hyperparameters["smoothTailHeadEmbeded"]:
        for smoothTailIteration in range(0, self._hyperparameters["smoothTailHeadEmbededNbOfIterations"]):
          points = self.__smoothTail(points, self._nbTailPoints, self._hyperparameters["smoothTailHeadEmbeded"])
      else:
        if not(self._hyperparameters["adjustHeadEmbededTracking"]):
          points   = self._nbTailPoints(points, self._nbTailPoints)
        else:
          nDist    = len(points[0]) - 1
          stepDist = 10 / nDist
          tab      = [int((i/9)*nDist) for i in range(0,10)]
          points   = [[points[0][i] for i in tab], [points[1][i] for i in tab]]


    if False:
      polynomialDegree = 5
      points = self.__smoothBasedOnCurvature(points, polynomialDegree)

    output = np.zeros((1, len(points[0]), 2))

    for idx, x in enumerate(points[0]):
      output[0][idx][0] = x
      output[0][idx][1] = points[1][idx]
    
    
    for idx, x in enumerate(points[0]):
      y = points[1][idx]
      cv2.circle(threshEyes, (int(x), int(y)), 50, (255, 255, 255), -1)
    if printDebbbbug:
      self._debugFrame(threshEyes, title='threshEyes1')
    threshEyes[0, :] = 255
    threshEyes[:, 0] = 255
    threshEyes[len(threshEyes)-1, :] = 255
    threshEyes[:, len(threshEyes[0])-1] = 255
    contours, hierarchy = cv2.findContours(threshEyes,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
    areas = np.array([cv2.contourArea(contour) for contour in contours])
    max1 = 0
    max2 = 0
    max1_index = -1
    max2_index = -1
    for index, value in enumerate(areas):
      if value >= 1000000:
        continue
      if value > max1:
        max2 = max1
        max2_index = max1_index
        max1 = value
        max1_index = index
      elif value > max2:
        max2 = value
        max2_index = index
    M1 = cv2.moments(contours[max1_index])
    if M1["m00"] != 0:
      cX1 = int(M1["m10"] / M1["m00"])
      cY1 = int(M1["m01"] / M1["m00"])
    else:
      cX1, cY1 = 0, 0
    M2 = cv2.moments(contours[max2_index])
    if M2["m00"] != 0:
      cX2 = int(M2["m10"] / M2["m00"])
      cY2 = int(M2["m01"] / M2["m00"])
    else:
      cX2, cY2 = 0, 0
    cX = (cX1 + cX2) / 2
    cY = (cY1 + cY2) / 2
    heading = self._calculateAngle(headPosition[0], headPosition[1], cX, cY)
    threshEyes = cv2.cvtColor(threshEyes, cv2.COLOR_GRAY2RGB)
    cv2.circle(threshEyes, (int(cX), int(cY)), 5, (0, 255, 0), -1)
    cv2.circle(threshEyes, (int(headPosition[0]), int(headPosition[1])), 5, (0, 255, 0), -1)
    if printDebbbbug:
      self._debugFrame(threshEyes, title='threshEyes')
    
    return output, heading

  def _headEmbededTailTrackFindMaxDepthForImage(self, frame):
    if not(("initialTailForMaxDepthCalculNotStraight" in self._hyperparameters) and self._hyperparameters["initialTailForMaxDepthCalculNotStraight"]):
      return math.sqrt((self._headPositionFirstFrame[0] - self._tailTipFirstFrame[0])**2 + (self._headPositionFirstFrame[1] - self._tailTipFirstFrame[1])**2)

    headEmbededParamTailDescentPixThreshStopInit = self._hyperparameters["headEmbededParamTailDescentPixThreshStop"]
    self._hyperparameters["headEmbededParamTailDescentPixThreshStop"] = 256

    x = self._headPositionFirstFrame[0]
    y = self._headPositionFirstFrame[1]

    steps   = self._hyperparameters["step"]
    nbList  = 10

    initialImage = frame.copy()
    
    #####
    # frame = cv2.GaussianBlur(frame, (5, 5), 0)
    # frame = cv2.Canny(frame, threshold1=50, threshold2=50) #threshold1=100, threshold2=200)
    #####
    
    back = frame.copy()
    back[:, :] = np.median(back[:3, :], axis=0)
    putToWhite = ( frame.astype('int32') >= (back.astype('int32') - 0) )
    frame[putToWhite] = 255
    ret, frame = cv2.threshold(frame.copy(), 254, 255, cv2.THRESH_BINARY)
    
    if applyDilateErode:
      erodeDilateSize = 25
      frame = cv2.erode(frame, np.ones((erodeDilateSize, erodeDilateSize), np.uint8), iterations=1)
      frame = cv2.dilate(frame, np.ones((erodeDilateSize, erodeDilateSize), np.uint8), iterations=1)
    
    #####
    
    if printDebbbbug:
      self._debugFrame(frame, title='Test1')
    
    # homogeneous = frame.copy()
    # homogeneous[:, :] = np.median(frame) - maxThresholdMinus #maxThreshold
    
    # frame = np.maximum(frame, homogeneous)
    
    if printDebbbbug:
      self._debugFrame(frame, title='Test2')
    
    gaussian_blur = self._hyperparameters["headEmbededParamGaussianBlur"]

    frame = cv2.GaussianBlur(frame, (gaussian_blur, gaussian_blur), 0)

    angle = self._calculateAngle(x, y, self._tailTipFirstFrame[0], self._tailTipFirstFrame[1])

    points = np.zeros((2, 0))
    
    if printDebbbbug:
      self._debugFrame(frame, title='Test3')

    (points, lastFirstTheta2) = self.__findNextPoints(0,x,y,frame,points,angle,self._hyperparameters["headEmbededTailTrackFindMaxDepthInitialMaxDepth"],steps,nbList,initialImage, self._hyperparameters["debugHeadEmbededFindNextPoints"])

    distToTip        = np.full((200),10000)
    distToBase       = np.full((200),10000)
    curTailLengthTab = np.full((200),10000)
    curTailLength  = 0

    k = 0
    distToTip[k]        = abs(math.sqrt((points[0,k]-self._tailTipFirstFrame[0])**2 + (points[1,k]-self._tailTipFirstFrame[1])**2))
    distToBase[k]       = abs(math.sqrt((points[0,k] - x)**2 + (points[1,k] - y)**2))
    curTailLength       = abs(math.sqrt((points[0,k] - x)**2 + (points[1,k] - y)**2))
    curTailLengthTab[k] = curTailLength

    k = 1
    distFromHeadToTip = abs(math.sqrt((x-self._tailTipFirstFrame[0])**2 + (y-self._tailTipFirstFrame[1])**2))
    while (curTailLength < 1.5*distFromHeadToTip) and (k < len(points[0])-1):
      curTailLength = curTailLength + abs(math.sqrt((points[0,k]-points[0,k-1])**2 + (points[1,k]-points[1,k-1])**2))
      distToTip[k]  = abs(math.sqrt((points[0,k]-self._tailTipFirstFrame[0])**2 + (points[1,k]-self._tailTipFirstFrame[1])**2))
      distToBase[k] = abs(math.sqrt((points[0,k] - x)**2 + (points[1,k] - y)**2))
      curTailLengthTab[k] = curTailLength
      k = k + 1

    minDistToTip    = 1000000
    indMinDistToTip = 0
    for idx, dist in enumerate(distToTip):
      if dist < minDistToTip:
        minDistToTip = dist
        indMinDistToTip = idx

    self._hyperparameters["headEmbededParamTailDescentPixThreshStop"] = headEmbededParamTailDescentPixThreshStopInit

    pathFactor = curTailLengthTab[indMinDistToTip] / distToBase[indMinDistToTip]

    return (math.sqrt((x - self._tailTipFirstFrame[0])**2 + (y - self._tailTipFirstFrame[1])**2)* pathFactor)



import struct
import csv
import mmap
import datetime
import array
import logging
import os
import csv
import glob
import time
import configparser
import numpy as np
import cv2
from pathlib import Path
import platform
import tifffile as tiff

class ZzVideoReading():

  def __init__(self, videoPath):
  
    seq_path  = videoPath

    cfg = configparser.ConfigParser()
    cfg.read(seq_path)
    
    width = int(cfg.get('Sequence Settings', 'Width'))
    height = int(cfg.get('Sequence Settings', 'Height'))
    bpp = int(cfg.get('Sequence Settings', 'BytesPerPixel'))
    num_images = cfg.get('Sequence Settings', 'Number of files')
    bin_file = cfg.get('Sequence Settings', 'Bin File')
    sqb_path = seq_path.replace('.seq', '.sqb')
    
    pathstr = os.path.dirname(seq_path)
    
    f = open(sqb_path, 'rb')
    
    self.sqb_path = sqb_path
    self.width = width
    self.height = height
    self.bpp = bpp
    self.num_images = int(num_images)
    self.bin_file = bin_file
    self.sqb_path = sqb_path
    self.pathstr = pathstr
    self.f = f
    self.lastFrameRead = -1
  
  def get(self, idOfInfoRequested):
    
    if idOfInfoRequested == 1:
      return self.lastFrameRead + 1
    elif idOfInfoRequested == 3:
      return self.width
    elif idOfInfoRequested == 4:
      return self.height
    elif idOfInfoRequested == 7:
      return self.num_images
    elif idOfInfoRequested == 5:
      return 24 # fake input video fps
  
  def isOpened(self):
    return True
  
  def release(self):
    del self
    
  def read(self):
    
    if self.lastFrameRead + 1 < self.num_images:
      
      try:
        offset = struct.unpack('i', self.f.read(4))
        padding = self.f.read(4)
        timestamp = struct.unpack('d', self.f.read(8))
        binfile = struct.unpack('i', self.f.read(4))
        padding = self.f.read(4)
      except:
        print("Problem with Hiris video format", self.lastFrameRead + 1)
        return [False, []]
      
      bin_path = os.path.join("%s" % (self.pathstr), "%s%0.5d.bin" % (self.bin_file, binfile[0]))
      
      if not(os.path.exists(bin_path)):
        print("Hiris video format: frame not found:", self.lastFrameRead + 1)
        return [False, []]
      
      f_bin = open(bin_path, 'rb')
      f_bin.seek(offset[0], os.SEEK_SET)
      
      bytes = f_bin.read(self.height*self.width*self.bpp)
      
      if self.bpp == 2:
        buffer = np.frombuffer(bytes, dtype=np.uint16)
      else:
        buffer = np.frombuffer(bytes, dtype=np.uint8)
      
      
      if len(buffer) == 3*self.height*self.width:
        nparr2 = buffer.reshape(self.height, self.width, 3)
      else:
        nparr2 = buffer.reshape(self.height, self.width)
        nparr2 = cv2.cvtColor(nparr2, cv2.COLOR_GRAY2RGB)
      
      self.lastFrameRead = self.lastFrameRead + 1
      
      return [True, nparr2]
    
    else:
      
      return [False, []]
  
  
  def set(self, propToChange, numImage):
    
    if propToChange == 1:
      
      if numImage <= 0:
        
        self.f.close()
        self.f = open(self.sqb_path, 'rb')
        self.lastFrameRead = -1
      
      elif numImage >= self.num_images:
        
        self.lastFrameRead = int(self.num_images)
        
      elif numImage > self.lastFrameRead:
        
        justToMove = self.f.read(24 * (int(numImage) - int(self.lastFrameRead) - 1))
        self.lastFrameRead = numImage - 1  
        
      elif numImage <= self.lastFrameRead:
        
        self.f.close()
        self.f = open(self.sqb_path, 'rb')
        
        justToMove = self.f.read(24 * int(numImage))
        self.lastFrameRead = numImage - 1


class tifReading():

  def __init__(self, videoPath):
    self.tif = tiff.TiffFile(videoPath)
    self.curPage    = 0
    self.num_images = len(self.tif.pages)
    firsPage = self.tif.pages[self.curPage]
    firsPage = firsPage.asarray()
    self.width  = firsPage.shape[0]
    self.height = firsPage.shape[1]
  
  def get(self, idOfInfoRequested):
    if idOfInfoRequested == 1:
      return self.curPage
    elif idOfInfoRequested == 3:
      return self.width
    elif idOfInfoRequested == 4:
      return self.height
    elif idOfInfoRequested == 7:
      return self.num_images
    elif idOfInfoRequested == 5:
      return 24 # fake input video fps
  
  def isOpened(self):
    if len(self.tif.pages):
      return True
    else:
      return False
    
  def release(self):
    self.tif.pages = []
    
  def read(self):
    if self.curPage < len(self.tif.pages):
      page = self.tif.pages[self.curPage]
      frame_data = page.asarray()
      if frame_data.dtype != 'uint8':
        frame_data = (frame_data / frame_data.max() * 255).astype('uint8')
      if not(len(frame_data.shape) == 3 and frame_data.shape[2] == 3):
        frame_data = cv2.cvtColor(frame_data, cv2.COLOR_GRAY2BGR)
      self.curPage += 1
      return [True, frame_data]
    else:
      return [False, []]
  
  def set(self, propToChange, numImage):
    if propToChange == 1:
      self.curPage = numImage


class eventBasedReading():
  
  def __init__(self, videoPath, hyperparameters=0):
    
    from metavision_core.event_io.raw_reader import RawReader
    
    self.record_raw = RawReader(videoPath.replace(".bias", ".raw"))
    self.height, self.width = self.record_raw.get_size()
    self.curPage    = 0
    
    self.im  = np.zeros((self.height, self.width))
    self.im[:, :] = 255
    self.lastIm = np.zeros((self.height, self.width))
    self.lastIm[:, :] = 255
    
    if hyperparameters:
      self.delta_t_toLoad            = hyperparameters["delta_t_toLoad"]
      self.nbPixelsAddAtEachFrame    = hyperparameters["nbPixelsAddAtEachFrame"]
      self.maxSumAllPixelToKeepImage = hyperparameters["maxSumAllPixelToKeepImage"]
      self.fps = 1 / (self.delta_t_toLoad / 1000000)
      
      events = self.record_raw.load_n_events(5000)
      firstEventTime = events[0][3]
      while not(self.record_raw.is_done()):
        events = self.record_raw.load_n_events(5000)
        lastEventTime = events[len(events) - 1][3]
      secondsLenght = (lastEventTime - firstEventTime) / (1000 * 1000)
      
      nbFrames = int(secondsLenght * self.fps)
      self.nbFrames = nbFrames
      print("nbFrames:", nbFrames, "; secondsLenght:", secondsLenght)
      self.record_raw.reset()
    else:
      self.nbFrames = -1
  
  def get(self, idOfInfoRequested):
    if idOfInfoRequested == 1:
      return self.curPage
    elif idOfInfoRequested == 3:
      return self.width
    elif idOfInfoRequested == 4:
      return self.height
    elif idOfInfoRequested == 7:
      return self.nbFrames
    elif idOfInfoRequested == 5:
      return self.fps
  
  def isOpened(self):
    if self.record_raw:
      return (not self.record_raw.is_done())
    else:
      return False
  
  def release(self):
    self.record_raw.reset()

  # Retrieve only latest events
  # def read(self):
    # if not self.record_raw.is_done():
      # events = self.record_raw.load_delta_t(self.delta_t_toLoad)
      # self.lastIm = np.zeros((self.height, self.width))
      # self.lastIm[:, :] = 255
      # self.im = self.lastIm
      # events2 = np.array([events['x'], events['y'], events['p']])
      # self.im[events2[1], events2[0]] = events2[2]
      # return [True, self.im.astype('uint8')]
    # else:
      # return [False, []]
  
  # Time surface, all events
  def read(self):
    if not self.record_raw.is_done():
      events = self.record_raw.load_delta_t(self.delta_t_toLoad)
      self.im = self.lastIm + self.nbPixelsAddAtEachFrame
      events2 = np.array([events['x'], events['y'], events['p']])
      self.im[events2[1], events2[0]] = 0 #events2[2]
      self.im[self.im > 255] = 255
      summm = np.sum(self.im) / (self.height * self.width)
      if summm > self.maxSumAllPixelToKeepImage:
        self.im = self.lastIm
        self.im[events2[1], events2[0]] = 0 #events2[2]
      self.lastIm = self.im
      return [True, self.im.astype('uint8')]
    else:
      return [False, []]

  # Time surface, choosing only positive or negative events
  # def read(self):
    # if not self.record_raw.is_done():
      # events = self.record_raw.load_delta_t(self.delta_t_toLoad)
      # self.im = self.lastIm + self.nbPixelsAddAtEachFrame
      # events2 = np.array([events['x'], events['y'], events['p']])
      # if len(events2[0]) > 0:
        # events2 = events2[:, events2[2] != 1]
      # self.im[events2[1], events2[0]] = 0 #events2[2]
      # self.im[self.im > 255] = 255
      # summm = np.sum(self.im) / (self.height * self.width)
      # if summm > self.maxSumAllPixelToKeepImage:
        # self.im = self.lastIm
        # self.im[events2[1], events2[0]] = 0 #events2[2]
      # self.lastIm = self.im
      # return [True, self.im.astype('uint8')]
    # else:
      # return [False, []]
  
  def set(self, propToChange, numImage):
    if propToChange == 1:
      print("setting image position")
      if numImage <= self.curPage:
        self.record_raw.reset()
      self.curPage = numImage
      self.record_raw.seek_time((numImage / self.fps) * 1000 * 1000)


def VideoCapture(videoPath, hyperparameters=0):
  
  if '.seq' in videoPath:
    
    zzVidCapture = ZzVideoReading(videoPath)
    
    return zzVidCapture
  
  elif '.sqb' in videoPath:
    
    zzVidCapture = ZzVideoReading(videoPath.replace('.sqb', '.seq'))
    
    return zzVidCapture
  
  elif ('.tif' in videoPath) or ('.tiff' in videoPath):
  
    zzVidCapture = tifReading(videoPath)
    
    return zzVidCapture

  elif '.bias' in videoPath:
    
    zzVidCapture = eventBasedReading(videoPath, hyperparameters)
    
    return zzVidCapture
    
  else:
    
    return cv2.VideoCapture(videoPath)



import numpy as np

from ._calculateAndStoreTailAngleHeatmap import calculateAndStoreTailAngleHeatmap
from ._openResultsFile import openResultsFile


def getTailAngleHeatmapPerBout(videoName: str, numWell: int, numAnimal: int, numBout: int) -> tuple:
  with openResultsFile(videoName, 'r+') as results:
    boutsPath = f'dataForWell{numWell}/dataForAnimal{numAnimal}/listOfBouts'
    if boutsPath not in results:
      raise ValueError(f"bouts not found for animal {numAnimal} in well {numWell}")
    boutsGroup = results[boutsPath]
    boutGroup = boutsGroup[f'bout{numBout}']
    firstFrame = results.attrs['firstFrame']
    start = boutGroup.attrs['BoutStart'] - firstFrame
    end = boutGroup.attrs['BoutEnd'] - firstFrame + 1
    dataGroup = results[f'dataForWell{numWell}/dataForAnimal{numAnimal}/dataPerFrame']
    if 'tailAngleHeatmap' in dataGroup:
      tailAngleHeatmap = dataGroup['tailAngleHeatmap']
    else:
      print(f'calculating and storing tail angle heatmap for all bouts for well {numWell}, animal {numAnimal}')
      tailAngleHeatmap = calculateAndStoreTailAngleHeatmap(results, dataGroup, boutsGroup)
    return [tailAngleHeatmap[col][start:end] for col in tailAngleHeatmap.dtype.names], boutGroup.attrs['BoutStart'], dataGroup['TailLength'][0]



from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.utilities import appendPoint, distBetweenThetas, assignValueIfBetweenRange, calculateAngle
import zebrazoom.code.util as util
import numpy as np
import math
import cv2

def tailTrackFindNextPoint(depth, headPosition, frame, points, angle, maxDepth, steps, nbList, hyperparameters, debug, lenX, lenY, secondTry=False):
  
  x = headPosition[0]
  y = headPosition[1]
  
  pixSurMax = hyperparameters["headEmbededParamTailDescentPixThreshStop"]
  
  pixTotList = []
  
  distSubsquentPoints = 0.000001
  pixSur = 0
  
  while (distSubsquentPoints > 0 and depth < maxDepth and ((pixSur < pixSurMax) or (depth < hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth))):
  
    if depth == 0 and angle == -99999: # First frame, first segment
      thetaDiffAccept = 3.6
    else:
      if (depth < hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth):
        thetaDiffAccept = hyperparameters["thetaDiffAccept"]
      else:
        if not("authorizedRelativeLengthTailEnd2" in hyperparameters) or (depth < hyperparameters["authorizedRelativeLengthTailEnd2"]*maxDepth):
          thetaDiffAccept = hyperparameters["thetaDiffAcceptAfterAuthorizedRelativeLengthTailEnd"]
          nbList = hyperparameters["nbListAfterAuthorizedRelativeLengthTailEnd"]
        else:
          thetaDiffAccept = hyperparameters["thetaDiffAcceptAfterAuthorizedRelativeLengthTailEnd2"]
          nbList = hyperparameters["nbListAfterAuthorizedRelativeLengthTailEnd2"]
          

    pixTotMax = 1000000
    maxTheta  = angle

    l = [i*(math.pi/nbList) for i in range(0,2*nbList) if distBetweenThetas(i*(math.pi/nbList), angle) < thetaDiffAccept]

    for step in steps:
      if (step < maxDepth - depth) or (step == steps[0]):
        for theta in l:
          xNew = assignValueIfBetweenRange(int(x + step * (math.cos(theta))), 0, lenX)
          yNew = assignValueIfBetweenRange(int(y + step * (math.sin(theta))), 0, lenY)
          pixTot = frame[yNew][xNew]
          if (pixTot < pixTotMax):
            pixTotMax = pixTot
            maxTheta = theta
            xTot = xNew
            yTot = yNew
    
    pixTotList.append(pixTotMax)
    
    if False:
      w = 4
      ym = yTot - w
      yM = yTot + w
      xm = xTot - w
      xM = xTot + w
      if ym < 0:
        ym = 0
      if xm < 0:
        xm = 0
      if yM > lenY + 1:
        yM = lenY + 1
      if xM > lenX + 1:
        xM = lenX + 1
      pixSur = np.min(frame[ym:yM, xm:xM])
    else:
      pixSur = frame[yTot, xTot]

    # Calculates distance between new and old point
    distSubsquentPoints = math.sqrt((xTot - x)**2 + (yTot - y)**2)
    
    if depth + distSubsquentPoints < maxDepth and ((pixSur < pixSurMax) or (depth < hyperparameters["authorizedRelativeLengthTailEnd"]*maxDepth)):
      points = appendPoint(xTot, yTot, points)
    else:
      vectX = xTot - x
      vectY = yTot - y
      xTot  = int(x + (maxDepth / (depth + distSubsquentPoints)) * vectX)
      yTot  = int(y + (maxDepth / (depth + distSubsquentPoints)) * vectY)
      points = appendPoint(xTot, yTot, points)
    
    if debug:
      frameDisplay = frame.copy()
      cv2.circle(frameDisplay, (xTot, yTot), 1, (0,0,0),   -1)
      util.showFrame(frameDisplay, title="HeadEmbeddedTailTracking")
    
    newTheta = calculateAngle(x,y,xTot,yTot)
    
    angle = newTheta
    if depth == 0:
      lastFirstTheta = angle
    depth = depth + distSubsquentPoints
    x = xTot
    y = yTot
  
  lenPoints = len(points[0]) - 1
  if points[0, lenPoints-1] == points[0, lenPoints] and points[1, lenPoints-1] == points[1, lenPoints]:
    points = points[:, :len(points[0])-1]
  
  if debug:
    print(np.median(pixTotList), np.mean(pixTotList), pixTotList)
  
  medianPixTotList = np.median(pixTotList)
  if "tries4rotationsCombination" in hyperparameters and hyperparameters["tries4rotationsCombination"] and len(pixTotList) <= 2:
    medianPixTotList += 25
  
  if hyperparameters["maximumMedianValueOfAllPointsAlongTheTail"] and medianPixTotList > hyperparameters["maximumMedianValueOfAllPointsAlongTheTail"] and not(secondTry):
    # print("second try")
    (points2, lastFirstTheta2, medianPixTotList2) = tailTrackFindNextPoint(0, headPosition, frame, np.zeros((2, 0)), (lastFirstTheta + math.pi) % (2 * math.pi), maxDepth, steps, nbList, hyperparameters, debug, lenX, lenY, True)
    if medianPixTotList2 < medianPixTotList:
      return (points2, lastFirstTheta2, medianPixTotList2)
  
  return (points, lastFirstTheta, medianPixTotList)


import math

import numpy as np

from .getDataPerBout import getDataPerBout


def _distanceBetweenPointAndLine(p1, p2, p3, pixelSize):  # p1 and p2 define the line, p3 is the point
  x0, y0 = p3
  x1, y1 = p1
  x2, y2 = p2
  return -pixelSize * ((y2 - y1) * x0 - (x2 - x1) * y0 + x2 * y1 - y2 * x1) / math.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)


def getSingleFrameDistanceToAxis(videoName: str, pixelSize):
  heading = getDataPerBout(videoName, 0, 0, 0, 'Heading')[0]
  x1, y1 = getDataPerBout(videoName, 0, 0, 0, 'HeadPos')[0]
  x2 = x1 + 1000 * math.cos(heading)
  y2 = y1 + 1000 * math.sin(heading)
  return [_distanceBetweenPointAndLine((x1, y1), (x2, y2), pos, pixelSize) for pos in zip(getDataPerBout(videoName, 0, 0, 0, 'TailPosX')[0], getDataPerBout(videoName, 0, 0, 0, 'TailPosY')[0])]



from typing import Tuple

from ._openResultsFile import openResultsFile


def getFirstAndLastFrame(videoName: str) -> Tuple[int, int]:
  with openResultsFile(videoName, 'r') as results:
    return results.attrs['firstFrame'], results.attrs['lastFrame']



import numpy as np

from ._openResultsFile import openResultsFile


def deleteAutomaticallyCalculatedKinematicParameters(videoName: str):
  
  with openResultsFile(videoName, 'a') as results:

    for j in range(len(results["dataForWell0"])):
      del results["dataForWell0/dataForAnimal" + str(j) + "/kinematicParametersPerBout/"]
      for i in range(len(results["dataForWell0/dataForAnimal" + str(j) + "/listOfBouts/"].keys())):
        del results["dataForWell0/dataForAnimal" + str(j) + "/listOfBouts/bout" + str(i) + "/additionalKinematicParametersPerBout/"]
  


import re


_ALPHABET_REGEX = re.compile(r'[^a-zA-Z]+')
_SORT_PRIORITY = {'wt': 0, 'ctrl': 0, 'ctrls': 0, 'controls': 0, 'control': 0, 'wildtype': 0, 'wildtypes': 0,
                  'het': 1, 'heterozygote': 1, 'heterozygotes': 1,
                  'mut': 2, 'mutant': 2, 'mutants': 2}


def sortGenotypes(genotypes):
  def key(genotype):
    if not isinstance(genotype, str):
      return float('inf'), genotype
    return _SORT_PRIORITY.get(re.sub(_ALPHABET_REGEX, '', genotype).casefold(), float('inf')), genotype
  return sorted(genotypes, key=key)


import cv2

from zebrazoom.code.adjustHyperparameters import adjustFreelySwimTrackingParams, adjustFreelySwimTrackingAutoParams
from zebrazoom.code.tracking.fasterMultiprocessing2 import FasterMultiprocessing2
from zebrazoom.code.tracking import register_tracking_method
from ._base import BaseGUITrackingMethod


class GUIFasterMultiprocessing2(FasterMultiprocessing2, BaseGUITrackingMethod):
  def _adjustParameters(self, i, back, frame, initialCurFrame, widgets):
    if self._hyperparameters["adjustFreelySwimTracking"]:
      i, widgets = adjustFreelySwimTrackingParams(self._nbTailPoints, i, self._firstFrame, self._trackingHeadTailAllAnimalsList[self._firstWell], self._trackingHeadingAllAnimalsList[self._firstWell], frame, frame, self._hyperparameters, widgets)
    elif self._hyperparameters["adjustFreelySwimTrackingAutomaticParameters"]:
      # Preparing image to show
      if self._hyperparameters["recalculateForegroundImageBasedOnBodyArea"] and "minPixelDiffForBackExtractBody" in self._hyperparameters:
          minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtractBody"]
      else:
        if self._hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"] and "minPixelDiffForBackExtractHead" in self._hyperparameters:
          minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtractHead"]
          del self._hyperparameters["minPixelDiffForBackExtractHead"] # Not sure why this is necessary: need to check the code to make sure there isn't a bug somewhere
        else:
          minPixelDiffForBackExtract = self._hyperparameters["minPixelDiffForBackExtract"]
      curFrame = initialCurFrame.copy()
      putToWhite = (curFrame.astype('int32') >= (back.astype('int32') - minPixelDiffForBackExtract) )
      curFrame[putToWhite] = 255
      ret, frame2 = cv2.threshold(curFrame, self._hyperparameters["thresholdForBlobImg"], 255, cv2.THRESH_BINARY)
      # Showing current image and waiting for next parameter/frame change
      i, widgets = adjustFreelySwimTrackingAutoParams(self._nbTailPoints, i, self._firstFrame, self._trackingHeadTailAllAnimalsList[self._firstWell], self._trackingHeadingAllAnimalsList[self._firstWell], frame, frame2, self._hyperparameters, widgets)
      # Puts hyperparameters values to accepted values
      self._hyperparameters["recalculateForegroundImageBasedOnBodyArea"] = 0 if self._hyperparameters["recalculateForegroundImageBasedOnBodyArea"] < 0.5 else 1
      if self._hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"] < 0:
        self._hyperparameters["adjustMinPixelDiffForBackExtract_nbBlackPixelsMax"] = 0
      if self._hyperparameters["minPixelDiffForBackExtract"] < 0:
        self._hyperparameters["minPixelDiffForBackExtract"] = 0
      else:
        if self._hyperparameters["minPixelDiffForBackExtract"] > 255:
          self._hyperparameters["minPixelDiffForBackExtract"] = 255
      self._hyperparameters["minPixelDiffForBackExtract"] = int(self._hyperparameters["minPixelDiffForBackExtract"])
    else:
      return None
    return i, widgets


register_tracking_method('fasterMultiprocessing2', GUIFasterMultiprocessing2)



import numpy as np
import json
import cv2
import zebrazoom.videoFormatConversion.zzVideoReading as zzVideoReading
import math
import json
import os

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QFileDialog, QProgressDialog

import zebrazoom.code.util as util


def chooseVideoToTroubleshootSplitVideo(self, controller):

  # Choosing video to split
  self.videoToTroubleshootSplitVideo, _ =  QFileDialog.getOpenFileName(self.window, "Select video", os.path.expanduser("~"), "All files(*)")
  if not self.videoToTroubleshootSplitVideo:
    return

  def beginningAndEndChosen():
    # Choosing directory to save sub-video
    firstFrame = self.configFile["firstFrame"]
    lastFrame = self.configFile["lastFrame"]
    self.configFile = configFile
    filename, _ = QFileDialog.getSaveFileName(controller.window, 'Choose where you want to save the sub-video.', os.path.join(os.path.dirname(self.videoToTroubleshootSplitVideo), 'subvideo.avi'), "AVI (*.avi)")
    if not filename:
      return

    # Extracting sub-video

    cap = zzVideoReading.VideoCapture(self.videoToTroubleshootSplitVideo)
    if not cap.isOpened():
      print("Error opening video stream or file")
      return
    cap.set(1, firstFrame)
    width = int(cap.get(3))
    height = int(cap.get(4))

    out = cv2.VideoWriter(filename, cv2.VideoWriter_fourcc('M','J','P','G'), 10, (width, height))
    # out = cv2.VideoWriter(filename, cv2.VideoWriter_fourcc('H','F','Y','U'), 10, (width, height))

    cancelled = False
    progressDialog = QProgressDialog("Saving video...", "Cancel", firstFrame, lastFrame, controller.window, Qt.WindowType.WindowTitleHint | Qt.WindowType.WindowCloseButtonHint)
    progressDialog.setWindowTitle('Saving Video')
    progressDialog.setAutoClose(False)
    progressDialog.setAutoReset(False)
    progressDialog.setWindowModality(Qt.WindowModality.ApplicationModal)
    progressDialog.setMinimumDuration(0)
    out = cv2.VideoWriter(filename, cv2.VideoWriter_fourcc(*'MJPG'), 10, (width, height))
    for frameIdx in range(firstFrame, lastFrame):
      progressDialog.setValue(frameIdx)
      progressDialog.setLabelText(f'Saving frame {frameIdx}...')
      ret, frame = cap.read()
      if not ret:
        print(f"Error reading at frame {frameIdx}")
        break
      frameHeight, frameWidth = frame.shape[:2]
      if frameHeight < height or frameWidth < width:
        xOffset = (width - frameWidth) // 2
        yOffset = (height - frameHeight) // 2
        copy = cv2.resize(frame, (height, width))
        copy[:] = 0
        copy[yOffset:yOffset+frameHeight, xOffset:xOffset+frameWidth] = frame
        frame = copy
      out.write(frame)
      if progressDialog.wasCanceled():
        cancelled = True
        break
    progressDialog.setLabelText('Saving video...')
    cap.release()
    out.release()
    progressDialog.close()
    if not cancelled:
      self.show_frame("VideoToTroubleshootSplitVideo")
    else:
      self.show_frame("ChooseVideoToTroubleshootSplitVideo")

  configFile = self.configFile.copy()
  # User input of beginning and end of subvideo
  util.chooseBeginningPage(self, self.videoToTroubleshootSplitVideo, "Choose where the beginning of your sub-video should be", "Ok, I want the sub-video to start at this frame!",
                           lambda: util.chooseEndPage(self, self.videoToTroubleshootSplitVideo, "Choose where the sub-video should end.", "Ok, I want the sub-video to end at this frame!", beginningAndEndChosen,
                                                      leftButtonInfo=("Go to the start page", lambda: setattr(self, 'configFile', configFile) or self.show_frame("StartPage"))),
                           leftButtonInfo=("Go to the start page", lambda: self.show_frame("StartPage")))



import numpy as np
import json
import cv2
import math
from zebrazoom.code.vars import getGlobalVariables
import json
import os
import subprocess
import sys
import pandas as pd
globalVariables = getGlobalVariables()

from PyQt5.QtWidgets import QFileDialog

import zebrazoom.code.paths as paths
import zebrazoom.code.util as util
from zebrazoom.dataAnalysis.datasetcreation.createDataFrame import createDataFrame
from zebrazoom.dataAnalysis.dataanalysis.applyClustering import applyClustering

from zebrazoom.dataAnalysis.datasetcreation.generatePklDataFileForVideo import generatePklDataFileForVideo


def openAnalysisFolder(self, homeDirectory, specificDirectory):
  dir_path = os.path.join(os.path.join(homeDirectory,'dataAnalysis'), specificDirectory)
  if sys.platform == "win32":
    os.startfile(dir_path)
  else:
    opener ="open" if sys.platform == "darwin" else "xdg-open"
    subprocess.call([opener, dir_path])


@util.showInProgressPage('Parameters calculation')
def boutClustering(self, controller, nbClustersToFind, FreelySwimming, HeadEmbeded, minNbBendForBoutDetect=3, nbVideosToSave=0, modelUsedForClustering=0, removeOutliers=False, frameStepForDistanceCalculation='4', removeBoutsContainingNanValuesInParametersUsedForClustering=True, forcePandasRecreation=0):

  if len(frameStepForDistanceCalculation) == 0:
    frameStepForDistanceCalculation = '4'
  
  if modelUsedForClustering == 0:
    modelUsedForClustering = 'KMeans'
  else:
    modelUsedForClustering = 'GaussianMixture'

  if len(minNbBendForBoutDetect) == 0:
    minNbBendForBoutDetect = 3
  else:
    minNbBendForBoutDetect = int(minNbBendForBoutDetect)

  if len(nbVideosToSave) == 0:
    nbVideosToSave = 0
  else:
    nbVideosToSave = int(nbVideosToSave)

  videoSaveFirstTenBouts = True if nbVideosToSave else False


  if len(controller.ZZoutputLocation) == 0:
    ZZoutputLocation = paths.getDefaultZZoutputFolder()
  else:
    ZZoutputLocation = controller.ZZoutputLocation

  # Creating the dataframe on which the clustering will be applied
  dataframeOptions = {
    'pathToExcelFile'                   : self.experimentOrganizationExcelFileAndFolder, # os.path.join(cur_dir_path, os.path.join('dataAnalysis', 'experimentOrganizationExcel')),
    'fileExtension'                     : '.' + self.experimentOrganizationExcel.split(".")[1],
    'resFolder'                         : os.path.join(paths.getDataAnalysisFolder(), 'data'),
    'nameOfFile'                        : self.experimentOrganizationExcel.split(".")[0],
    'smoothingFactorDynaParam'          : 0,   # 0.001
    'nbFramesTakenIntoAccount'          : -1, #28,
    'numberOfBendsIncludedForMaxDetect' : -1,
    'minNbBendForBoutDetect'            : minNbBendForBoutDetect,
    'defaultZZoutputFolderPath'         : ZZoutputLocation,
    'tailAngleKinematicParameterCalculation' : 1,
    'getTailAngleSignMultNormalized'    : 1,
    'computeTailAngleParamForCluster'   : True,
    'computeMassCenterParamForCluster'  : False
  }
  if int(FreelySwimming):
    dataframeOptions['computeMassCenterParamForCluster'] = True
    
  generatePklDataFileForVideo(os.path.join(self.experimentOrganizationExcelFileAndFolder, self.experimentOrganizationExcel), ZZoutputLocation, frameStepForDistanceCalculation, forcePandasRecreation)
    
  [conditions, genotypes, nbFramesTakenIntoAccount, globParam] = createDataFrame(dataframeOptions, "", forcePandasRecreation, [])
  # Applying the clustering on this dataframe
  clusteringOptions = {
    'analyzeAllWellsAtTheSameTime' : 0, # put this to 1 for head-embedded videos, and to 0 for multi-well videos
    'pathToVideos' : ZZoutputLocation,
    'nbCluster' : int(nbClustersToFind),
    #'nbPcaComponents' : 30,
    'nbFramesTakenIntoAccount' : nbFramesTakenIntoAccount,
    'scaleGraphs' : True,
    'showFigures' : False,
    'useFreqAmpAsym' : False,
    'useAngles' : False,
    'useAnglesSpeedHeadingDisp' : False,
    'useAnglesSpeedHeading' : False,
    'useAnglesSpeed' : False,
    'useAnglesHeading' : False,
    'useAnglesHeadingDisp' : False,
    'useFreqAmpAsymSpeedHeadingDisp' : False,
    'videoSaveFirstTenBouts' : videoSaveFirstTenBouts,
    'globalParametersCalculations' : True,
    'nbVideosToSave' : nbVideosToSave,
    'resFolder' : os.path.join(paths.getDataAnalysisFolder(), 'data/'),
    'nameOfFile' : self.experimentOrganizationExcel.split(".")[0],
    'modelUsedForClustering' : modelUsedForClustering,
    'removeOutliers'         : removeOutliers,
    'removeBoutsContainingNanValuesInParametersUsedForClustering' : removeBoutsContainingNanValuesInParametersUsedForClustering
  }
  if int(FreelySwimming):
    clusteringOptions['useAnglesSpeedHeading'] = True
    # clusteringOptions['useAngleAnd3GlobalParameters'] = True
    # clusteringOptions['useFreqAmpAsym'] = True
  if int(HeadEmbeded):
    clusteringOptions['useAngles'] = True
  # Applies the clustering
  [allBouts, classifier] = applyClustering(clusteringOptions, 0, os.path.join(paths.getDataAnalysisFolder(), 'resultsClustering/'), self.ZZoutputLocation)
  # Saves the classifier
  controller.show_frame("AnalysisOutputFolderClustering")




from zebrazoom.code.tracking.customTrackingImplementations.fastFishTracking.utilities import calculateAngle, distBetweenThetas
import numpy as np
import queue
import math
import cv2

def detectMovementWithTrackedDataAfterTracking(self):

  trackingDataPerWell = self._trackingDataPerWell
  
  outputData = {} # Each element of this object will correspond to the tracking data of a particular well/tank/arena
    
  for wellNumber in range(0, len(self._wellPositions)):
    
    outputDataForWell = []
      
    if self._hyperparameters["detectBouts"]: # See below ("else") for no bout detection scenario
      
      nbFramesStepToAvoidNoise = self._hyperparameters["nbFramesStepToAvoidNoise"]
      minNbFramesForBoutDetect = self._hyperparameters["minNbFramesForBoutDetect"]
      
      # Finding frames with an instantaneous distance over a predifined threshold
      boutOccuring = [math.sqrt((trackingDataPerWell[wellNumber][i+nbFramesStepToAvoidNoise][0][0] - trackingDataPerWell[wellNumber][i][0][0])**2 + (trackingDataPerWell[wellNumber][i+nbFramesStepToAvoidNoise][0][1] - trackingDataPerWell[wellNumber][i][0][1])**2) > self._hyperparameters["minimumInstantaneousDistanceForBoutDetect"] for i in range(0, len(trackingDataPerWell[wellNumber])-nbFramesStepToAvoidNoise)]
      
      # Detecting bouts by finding long enough sequence of frames with high enough instantaneous distance
      boutCurrentlyOccuring = False
      boutFrameNumberStart  = -1
      for frameNumber, boutIsOccuring in enumerate(boutOccuring):
        if boutIsOccuring:
          if not(boutCurrentlyOccuring):
            boutFrameNumberStart = frameNumber
          boutCurrentlyOccuring = True
        else:
          if boutCurrentlyOccuring:
            boutCurrentlyOccuring = False
            if frameNumber - boutFrameNumberStart > minNbFramesForBoutDetect:
              # Saving information for each bout of movement detected
              boutOfMovement = {}
              boutOfMovement["AnimalNumber"]  = 0
              boutOfMovement["BoutStart"]     = boutFrameNumberStart
              boutOfMovement["BoutEnd"]       = frameNumber
              boutOfMovement["HeadX"]         = [trackingDataPerWell[wellNumber][i][0][0] for i in range(boutFrameNumberStart, frameNumber)]
              boutOfMovement["HeadY"]         = [trackingDataPerWell[wellNumber][i][0][1] for i in range(boutFrameNumberStart, frameNumber)]
              boutOfMovement["Heading"]                = [(calculateAngle(trackingDataPerWell[wellNumber][i][0][0], trackingDataPerWell[wellNumber][i][0][1], trackingDataPerWell[wellNumber][i][1][0], trackingDataPerWell[wellNumber][i][1][1]) + math.pi) % (2 * math.pi) for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["TailAngle_Raw"]          = [distBetweenThetas(calculateAngle(trackingDataPerWell[wellNumber][i][0][0], trackingDataPerWell[wellNumber][i][0][1], trackingDataPerWell[wellNumber][i][1][0], trackingDataPerWell[wellNumber][i][1][1]), calculateAngle(trackingDataPerWell[wellNumber][i][0][0], trackingDataPerWell[wellNumber][i][0][1], trackingDataPerWell[wellNumber][i][len(trackingDataPerWell[wellNumber][i])-1][0], trackingDataPerWell[wellNumber][i][len(trackingDataPerWell[wellNumber][i])-1][1])) for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["TailX_VideoReferential"] = [[trackingDataPerWell[wellNumber][i][j][0] for j in range(0, len(trackingDataPerWell[wellNumber][i]))] for i in range(boutFrameNumberStart, frameNumber+1)]
              boutOfMovement["TailY_VideoReferential"] = [[trackingDataPerWell[wellNumber][i][j][1] for j in range(0, len(trackingDataPerWell[wellNumber][i]))] for i in range(boutFrameNumberStart, frameNumber+1)]
              
              outputDataForWell.append(boutOfMovement)
    
    else: # No bout detction in this case
    
      lastFrame = len(trackingDataPerWell[wellNumber])
      boutOfMovement = {}
      boutOfMovement["AnimalNumber"]  = 0
      boutOfMovement["BoutStart"]     = 0
      boutOfMovement["BoutEnd"]       = len(trackingDataPerWell[wellNumber][0])
      boutOfMovement["HeadX"]         = [trackingDataPerWell[wellNumber][0][i][0][0] for i in range(0, lastFrame)]
      boutOfMovement["HeadY"]         = [trackingDataPerWell[wellNumber][0][i][0][1] for i in range(0, lastFrame)]
      boutOfMovement["Heading"]       = [(calculateAngle(trackingDataPerWell[wellNumber][0][i][0][0], trackingDataPerWell[wellNumber][0][i][0][1], trackingDataPerWell[wellNumber][0][i][1][0], trackingDataPerWell[wellNumber][0][i][1][1]) + math.pi) % (2 * math.pi) for i in range(0, lastFrame)]
      boutOfMovement["TailAngle_Raw"]          = [distBetweenThetas(calculateAngle(trackingDataPerWell[wellNumber][0][i][0][0], trackingDataPerWell[wellNumber][0][i][0][1], trackingDataPerWell[wellNumber][0][i][1][0], trackingDataPerWell[wellNumber][0][i][1][1]), calculateAngle(trackingDataPerWell[wellNumber][0][i][0][0], trackingDataPerWell[wellNumber][0][i][0][1], trackingDataPerWell[wellNumber][0][i][len(trackingDataPerWell[wellNumber][0][i])-1][0], trackingDataPerWell[wellNumber][0][i][len(trackingDataPerWell[wellNumber][0][i])-1][1])) for i in range(0, lastFrame)]
      boutOfMovement["TailX_VideoReferential"] = [[trackingDataPerWell[wellNumber][0][i][j][0] for j in range(0, len(trackingDataPerWell[wellNumber][0][i]))] for i in range(0, lastFrame)]
      boutOfMovement["TailY_VideoReferential"] = [[trackingDataPerWell[wellNumber][0][i][j][1] for j in range(0, len(trackingDataPerWell[wellNumber][0][i]))] for i in range(0, lastFrame)]
      outputDataForWell.append(boutOfMovement)
    
    # Saving all tracked bouts of movements for current frame
    outputData[wellNumber] = outputDataForWell
  
  return outputData